"""
Evaluate on ImageNet. Note that at the moment, training is not implemented (I am working on it).
that being said, evaluation is working.
"""

import argparse
import os
import random
import shutil
import time
import warnings
import PIL
import datetime

import torch
import torch.nn as nn
import torch.nn.parallel
import torch.backends.cudnn as cudnn
import torch.distributed as dist
import torch.optim
import torch.multiprocessing as mp
import torch.utils.data
import torch.utils.data.distributed
import torchvision.transforms as transforms
import torchvision.datasets as datasets
import torchvision.models as models
from torch.optim.lr_scheduler import LambdaLR

from .efficientnet_pytorch import EfficientNet,T_cls


def compute_lr_lambda(epoch, warmup_epochs, num_epochs, power, base_values, eta_min):
    if epoch < warmup_epochs:
        return (1.0 / (warmup_epochs + 1)) * (epoch + 1)

    factor = (1 - (epoch - warmup_epochs) / float(num_epochs - warmup_epochs)) ** power
    return [(base_lr - eta_min) * factor + eta_min for base_lr in base_values][0]

best_acc1=0
class modelEfficient:
    def __init__(self, params):
        self.params = params
        self.model_savedir = None
        # self.best_acc1 = 0

    class ClsPresetTrain:
        def __init__(self, imgsz, hflip_prob=0.5, mean=(0.485, 0.456, 0.406), std=(0.229, 0.224, 0.225)):

            trans = [T_cls.imgResize((imgsz[0], imgsz[1]))]
            if hflip_prob > 0:
                trans.append(T_cls.RandomHorizontalFlip(hflip_prob))
            trans.extend([
                T_cls.ToTensor(),
                T_cls.Normalize(mean=mean, std=std),
            ])
            self.transforms = T_cls.Compose(trans)

        def __call__(self, img, target=None):
            # 确保 img 和 target 都传递给变换
            if target is None:
                return self.transforms(img)
            else:
                # 将 img 和 target 作为元组传递给变换
                return self.transforms((img, target))

    class ClsPresetEval:
        def __init__(self, imgsz, mean=(0.485, 0.456, 0.406), std=(0.229, 0.224, 0.225)):
            # 根据 crop_size 是否存在来决定是否添加 T.RandomCrop
            trans=[T_cls.imgResize((imgsz[0], imgsz[1]))]
            trans.extend([
                T_cls.ToTensor(),
                T_cls.Normalize(mean=mean, std=std),
            ])
            self.transforms = T_cls.Compose(trans)

        def __call__(self, img, target=None):
            # 确保 img 和 target 都传递给变换
            if target is None:
                return self.transforms(img)
            else:
                # 将 img 和 target 作为元组传递给变换
                return self.transforms((img, target))

    def get_transform_cls(self,imgsz,train):
        return self.ClsPresetTrain(imgsz) if train else self.ClsPresetEval(imgsz)

    def train_Efficient(self):
        if self.params["seed"] is not None:
            random.seed(self.params["seed"])
            torch.manual_seed(self.params["seed"])
            cudnn.deterministic = True
            warnings.warn('You have chosen to seed training. '
                          'This will turn on the CUDNN deterministic setting, '
                          'which can slow down your training considerably! '
                          'You may see unexpected behavior when restarting '
                          'from checkpoints.')

        if self.params["device"] is not None:
            warnings.warn('You have chosen a specific GPU. This will completely '
                          'disable data parallelism.')

        if self.params["dist_url"] == "env://" and self.params["world_size"] == -1:
            self.params["world_size"] = int(os.environ["WORLD_SIZE"])

        self.params["distributed"] = self.params["world_size"] > 1 or self.params["multiprocessing_distributed"]

        ngpus_per_node = torch.cuda.device_count()
        if self.params["multiprocessing_distributed"]:
            # Since we have ngpus_per_node processes per node, the total world_size
            # needs to be adjusted accordingly
            self.params["world_size"] = ngpus_per_node * self.params["world_size"]
            # Use torch.multiprocessing.spawn to launch distributed processes: the
            # main_worker process function
            mp.spawn(self.main_worker, nprocs=ngpus_per_node, args=(ngpus_per_node, self.params))
        else:
            # Simply call main_worker function
            self.main_worker(self.params["device"], ngpus_per_node, self.params)
    def main_worker(self,gpu, ngpus_per_node, params_mw):
        global best_acc1
        params_mw["device"] = gpu
        if params_mw["device"] is not None:
            print("Use GPU: {} for training".format(params_mw["device"]))

        if params_mw["distributed"]:
            if params_mw["dist_url"] == "env://" and params_mw["rank"] == -1:
                params_mw["rank"] = int(os.environ["RANK"])
            if params_mw["multiprocessing_distributed"]:
                # For multiprocessing distributed training, rank needs to be the
                # global rank among all the processes
                params_mw["rank"]  =params_mw["rank"]  * ngpus_per_node + gpu
            dist.init_process_group(backend=params_mw["dist_backend"], init_method=params_mw["dist_url"],
                                    world_size=params_mw["world_size"], rank=params_mw["rank"])
        # create model
        if 'efficientnet' in params_mw["arch"]:  # NEW
            if params_mw["pretrained"]:
                if params_mw["pretrained_model"]:
                    model = EfficientNet.from_pretrained(params_mw["arch"], weights_path=params_mw["pretrained_model"],advprop=params_mw["advprop"])
                    print("=> using pre-trained model '{}'".format(params_mw["pretrained_model"]))
                else:
                    model = EfficientNet.from_pretrained(params_mw["arch"], advprop=params_mw["advprop"])
                    print("=> using pre-trained model '{}'".format(params_mw["arch"]))
            else:
                print("=> creating model '{}'".format(params_mw["arch"]))
                model = EfficientNet.from_name(params_mw["arch"])

        else:
            if params_mw["pretrained"]:
                if params_mw["pretrained_model"]:
                    print('Loading pretrained weights for {}'.format(params_mw["arch"]))
                    print("=> using pre-trained model '{}'".format(params_mw["pretrained_model"]))
                    model=torch.load(params_mw["pretrained_model"])['state_dict']
                else:
                    print("=> using pre-trained model '{}'".format(params_mw["arch"]))
                    model = models.__dict__[params_mw["arch"]](pretrained=True)
            else:
                print("=> creating model '{}'".format(params_mw["arch"]))
                model = models.__dict__[params_mw["arch"]](pretrained=False)

        if params_mw["distributed"]:
            # For multiprocessing distributed, DistributedDataParallel constructor
            # should always set the single device scope, otherwise,
            # DistributedDataParallel will use all available devices.
            if params_mw["device"] is not None:
                torch.cuda.set_device(params_mw["device"])
                model.cuda(params_mw["device"])
                # When using a single GPU per process and per
                # DistributedDataParallel, we need to divide the batch size
                # ourselves based on the total number of GPUs we have
                params_mw["batch_size"] = int(params_mw["batch_size"] / ngpus_per_node)
                params_mw["workers"] = int(params_mw["workers"] / ngpus_per_node)
                model = torch.nn.parallel.DistributedDataParallel(model, device_ids=[params_mw["device"]])
            else:
                model.cuda()
                # DistributedDataParallel will divide and allocate batch_size to all
                # available GPUs if device_ids are not set
                model = torch.nn.parallel.DistributedDataParallel(model)
        elif params_mw["device"] is not None:
            torch.cuda.set_device(params_mw["device"])
            model = model.cuda(params_mw["device"])
        else:
            # DataParallel will divide and allocate batch_size to all available GPUs
            if params_mw["arch"].startswith('alexnet') or params_mw["arch"].startswith('vgg'):
                model.features = torch.nn.DataParallel(model.features)
                model.cuda()
            else:
                model = torch.nn.DataParallel(model).cuda()

        # define loss function (criterion) and optimizer
        criterion = nn.CrossEntropyLoss().cuda(params_mw["device"])

        optimizer = torch.optim.SGD(model.parameters(), params_mw["lr"],
                                    momentum=params_mw["momentum"],
                                    weight_decay=params_mw["weight_decay"])

        # optionally resume from a checkpoint
        if params_mw["resume"]:
            if os.path.isfile(params_mw["resume_model"]):
                print("=> loading checkpoint '{}'".format(params_mw["resume_model"]))
                checkpoint = torch.load(params_mw["resume_model"])
                params_mw["start_epoch"] = checkpoint['epoch']
                best_acc1 = checkpoint['best_acc1']
                if params_mw["device"] is not None:
                    # best_acc1 may be from a checkpoint from a different GPU
                    best_acc1 = best_acc1.to(params_mw["device"])
                model.load_state_dict(checkpoint['state_dict'].state_dict())
                # optimizer.load_state_dict(checkpoint['optimizer'])


                modelsavedir=checkpoint['outdir']
                epochs_all=checkpoint['epoch_all']
                print("=> loaded checkpoint '{}' (epoch {})"
                      .format(params_mw["resume_model"], checkpoint['epoch']))
            else:
                print("=> no checkpoint found at '{}'".format(params_mw["resume_model"]))
        else:
            ##创建结果txt
            modelsavedir = self.make_childdir()
            epochs_all =params_mw["epochs"]

        results_file = modelsavedir + "/results.txt"

        cudnn.benchmark = True

        # Data loading code
        traindir = os.path.join(params_mw["datadir"], 'train')
        valdir = os.path.join(params_mw["datadir"], 'val')
        if params_mw["advprop"]:
            normalize = transforms.Lambda(lambda img: img * 2.0 - 1.0)
        else:
            normalize = transforms.Normalize(mean=[0.485, 0.456, 0.406],
                                             std=[0.229, 0.224, 0.225])
        if params_mw["ifefficient_defaultsize"]:
            if 'efficientnet' in params_mw["arch"]:
                image_size = EfficientNet.get_image_size(params_mw["arch"])
            else:
                image_size = params_mw["image_size"]
        else:
            image_size = params_mw["image_size"]
        if params_mw["crop_size"]:
            train_transforms =transforms.Compose([
                transforms.RandomResizedCrop(image_size),
                transforms.RandomHorizontalFlip(),
                transforms.ToTensor(),
                normalize,
            ])
            val_transforms = transforms.Compose([
                transforms.Resize(image_size, interpolation=PIL.Image.BICUBIC),
                transforms.CenterCrop(image_size),
                transforms.ToTensor(),
                normalize,
            ])
        else:
            train_transforms = self.get_transform_cls([image_size, image_size], train=True)
            val_transforms = self.get_transform_cls([image_size, image_size], train=False)
        #     train_transforms =transforms.Compose([
        #         T_cls.imgResize((image_size, image_size)),
        #         transforms.RandomHorizontalFlip(),
        #         transforms.ToTensor(),
        #         normalize,
        #     ])
        #     val_transforms = transforms.Compose([
        #         T_cls.imgResize((image_size, image_size)),
        #         transforms.CenterCrop(image_size),
        #         transforms.ToTensor(),
        #         normalize,
        #     ])
        train_dataset = datasets.ImageFolder(
            traindir,
            train_transforms)

        if params_mw["distributed"]:
            train_sampler = torch.utils.data.distributed.DistributedSampler(train_dataset)
        else:
            train_sampler = None

        train_loader = torch.utils.data.DataLoader(
            train_dataset, batch_size=params_mw["batch_size"], shuffle=(train_sampler is None),
            num_workers=params_mw["workers"], pin_memory=True, sampler=train_sampler)

        print('Using image size', image_size)

        val_loader = torch.utils.data.DataLoader(
            datasets.ImageFolder(valdir, val_transforms),
            batch_size=params_mw["batch_size"], shuffle=False,
            num_workers=params_mw["workers"], pin_memory=True)

        if params_mw["evaluate"]:
            res = self.validate(val_loader, model, criterion, params_mw)
            with open('res.txt', 'w') as f:
                print(res, file=f)
            return

        scheduler = self.create_lr_scheduler(
            optimizer,
            num_step=len(train_loader),
            epochs=epochs_all,
            method=params_mw["lr_scheduler"],  # 可以选择 "step" 或 "poly"
            params_alr=params_mw
        )
        if params_mw["resume"]:
            if os.path.isfile(params_mw["resume_model"]):
                optimizer.load_state_dict(checkpoint['optimizer'])
                scheduler.load_state_dict(checkpoint['lr_scheduler'])
                # scheduler.step()

        for epoch in range(params_mw["start_epoch"],epochs_all):
            if params_mw["distributed"]:
                train_sampler.set_epoch(epoch)
            # self.adjust_learning_rate(optimizer, epoch, params_mw)

            # train for one epoch
            result_progress_train=self.train(train_loader, model, criterion, optimizer, epoch,epochs_all, params_mw)
            scheduler.step()
            # evaluate on validation set
            result_top1,result_top5,result_progress_val = self.validate(val_loader, model, criterion, params_mw)
            acc1=result_top1.avg
            # result_progress_train=result_progress_train.print_write(len(train_loader))
            # result_progress_val=result_progress_val.print_write(len(val_loader))
            # remember best acc@1 and save checkpoint
            is_best = acc1 > best_acc1
            best_acc1 = max(acc1, best_acc1)
            ##################
            # write into txt
            with open(results_file, "a") as f:
                # 记录每个epoch对应的train_loss、lr以及验证集各指标
                # val_info = f"[epoch: {epoch}]\n" \
                #              f"accuracy: {acc1:.4f}\n"
                # f.write(val_info + "\n\n")
                val_info = f"{result_progress_train}\n" \
                           f"{result_progress_val}\n" \
                           f"* Acc@1 {result_top1.avg:.3f} Acc@5 {result_top5.avg:.3f}\n"
                f.write(val_info + "\n\n")
            #################
            if not params_mw["multiprocessing_distributed"] or (params_mw["multiprocessing_distributed"]
                    and params_mw["rank"] % ngpus_per_node == 0):
                # modelsavedir=self.make_childdir()
                self.model_savedir=self.save_checkpoint({
                    'epoch': epoch + 1,
                    'epoch_all': params_mw["epochs"],
                    'arch': params_mw["arch"],
                    'state_dict': model,
                    'best_acc1': best_acc1,
                    'optimizer' : optimizer.state_dict(),
                    'lr_scheduler':scheduler.state_dict(),
                    'outdir':modelsavedir,
                }, is_best,filename=modelsavedir)

    def train(self,train_loader, model, criterion, optimizer, epoch, epochs_all,params_t):
        batch_time = self.AverageMeter('Time', ':6.3f')
        data_time = self.AverageMeter('Data', ':6.3f')
        losses = self.AverageMeter('Loss', ':.4e')
        top1 = self.AverageMeter('Acc@1', ':6.2f')
        top5 = self.AverageMeter('Acc@5', ':6.2f')
        progress = self.ProgressMeter(len(train_loader), batch_time, data_time, losses, top1,
                                 top5, prefix="Epoch: [{}/{}]".format(epoch,epochs_all))

        # switch to train mode
        model.train()

        end = time.time()
        for i, (images, target) in enumerate(train_loader):
            # measure data loading time
            data_time.update(time.time() - end)

            if params_t['device'] is not None:
                images = images.cuda(params_t['device'], non_blocking=True)
            target = target.cuda(params_t['device'], non_blocking=True)

            # compute output
            output = model(images)
            loss = criterion(output, target)

            # measure accuracy and record loss
            acc1, acc5 = self.accuracy(output, target, topk=(1, 5))
            losses.update(loss.item(), images.size(0))
            top1.update(acc1[0], images.size(0))
            top5.update(acc5[0], images.size(0))

            # compute gradient and do SGD step
            optimizer.zero_grad()
            loss.backward()
            optimizer.step()

            # measure elapsed time
            batch_time.update(time.time() - end)
            end = time.time()

            # if i % params_t["traininfo_printepoch"] == 0:
            #     progress.print(i)
            if params_t["traininfo_printepoch"]:
                if i % params_t["traininfo_printepoch"] == 0:
                    progress.print(i)
                    progress_out=progress.print_write(i)
            else:
                if (i % (len(train_loader)-1) == 0) & i>0:
                    progress.print(i)
                    progress_out = progress.print_write(i)
        return progress_out

    def validate(self,val_loader, model, criterion, params_v):
        batch_time = self.AverageMeter('Time', ':6.3f')
        losses = self.AverageMeter('Loss', ':.4e')
        top1 = self.AverageMeter('Acc@1', ':6.2f')
        top5 = self.AverageMeter('Acc@5', ':6.2f')
        progress =self.ProgressMeter(len(val_loader), batch_time, losses, top1, top5,
                                 prefix='Test: ')

        # switch to evaluate mode
        model.eval()

        with torch.no_grad():
            end = time.time()
            for i, (images, target) in enumerate(val_loader):
                if params_v['device'] is not None:
                    images = images.cuda(params_v['device'], non_blocking=True)
                target = target.cuda(params_v['device'], non_blocking=True)

                # compute output
                output = model(images)
                loss = criterion(output, target)

                # measure accuracy and record loss
                acc1, acc5 = self.accuracy(output, target, topk=(1, 5))
                losses.update(loss.item(), images.size(0))
                top1.update(acc1[0], images.size(0))
                top5.update(acc5[0], images.size(0))

                # measure elapsed time
                batch_time.update(time.time() - end)
                end = time.time()

                # if i % params_v["traininfo_printepoch"] == 0:
                #     progress.print(i)
                if params_v["traininfo_printepoch"]:
                    if i % params_v["traininfo_printepoch"] == 0:
                        progress.print(i)
                        progress_out=progress.print_write(i)
                else:
                    if (i % (len(val_loader)-1) == 0) & i>0:
                        progress.print(i)
                        progress_out = progress.print_write(i)
            # TODO: this should also be done with the ProgressMeter
            print(' * Acc@1 {top1.avg:.3f} Acc@5 {top5.avg:.3f}'
                  .format(top1=top1, top5=top5))

        # return top1.avg
        return top1,top5,progress_out
    def save_checkpoint(self,state, is_best, filename='checkpoint.pth'):
        efficientmodel_savedir = filename + '/weights'
        if not os.path.exists(efficientmodel_savedir):
            os.mkdir(efficientmodel_savedir)
        lastpath=efficientmodel_savedir+ "/last.pth"
        bestpath=efficientmodel_savedir+ "/best.pth"
        torch.save(state, lastpath)
        if is_best or (not os.path.exists(bestpath)):
            shutil.copyfile(lastpath, bestpath)
        return bestpath
    def make_childdir(self):
        # 用来保存训练以及验证过程中信息
        ####创建训练输出子文件train####
        from ultralytics.utils.files import increment_path
        train_childdir=os.path.join(self.params["save_dir"],'Efficientrain')
        train_dir = str(increment_path(train_childdir, exist_ok=False)).replace("\\",'/')
        ######
        if not os.path.exists(self.params["save_dir"]):
            os.mkdir(self.params["save_dir"])
        if not os.path.exists(train_dir):
            os.mkdir(train_dir)
        return train_dir

    def export_model(self,export_modeldict):
        model=torch.load(export_modeldict, map_location='cpu')['state_dict']
        return model

    class AverageMeter(object):
        """Computes and stores the average and current value"""
        def __init__(self, name, fmt=':f'):
            self.name = name
            self.fmt = fmt
            self.reset()

        def reset(self):
            self.val = 0
            self.avg = 0
            self.sum = 0
            self.count = 0

        def update(self, val, n=1):
            self.val = val
            self.sum += val * n
            self.count += n
            self.avg = self.sum / self.count

        def __str__(self):
            fmtstr = '{name} {val' + self.fmt + '} ({avg' + self.fmt + '})'
            return fmtstr.format(**self.__dict__)

    class ProgressMeter(object):
        def __init__(self, num_batches, *meters, prefix=""):
            self.batch_fmtstr = self._get_batch_fmtstr(num_batches)
            self.meters = meters
            self.prefix = prefix

        def print(self, batch):
            # entries = [self.prefix + self.batch_fmtstr.format(batch)]
            entries = [self.prefix + self.batch_fmtstr.format(batch+1)]
            entries += [str(meter) for meter in self.meters]
            print('\t'.join(entries))

        def print_write(self, batch):
            # entries = [self.prefix + self.batch_fmtstr.format(batch)]
            entries = [self.prefix + self.batch_fmtstr.format(batch + 1)]
            entries += [str(meter) for meter in self.meters]
            write_str='\t'.join(entries)
            return write_str

        def _get_batch_fmtstr(self, num_batches):
            num_digits = len(str(num_batches // 1))
            fmt = '{:' + str(num_digits) + 'd}'
            return '[' + fmt + '/' + fmt.format(num_batches) + ']'

    class PolyLRRatio(LambdaLR):
        def __init__(self, optimizer, eta_min=0, eta_min_ratio=None, power=1.0, num_epochs=100, warmup_epochs=0,
                     last_epoch=-1):
            self.eta_min = eta_min
            self.eta_min_ratio = eta_min_ratio
            self.power = power
            self.num_epochs = num_epochs
            self.warmup_epochs = warmup_epochs
            self.base_values = [group['lr'] for group in optimizer.param_groups]
            self.last_epoch = last_epoch
            super().__init__(optimizer, self._get_lr_lambda(), last_epoch)

        def _get_lr_lambda(self):
            return lambda epoch: compute_lr_lambda(epoch, self.warmup_epochs, self.num_epochs, self.power,
                                                   self.base_values, self.eta_min)

        def _get_value(self):
            param_groups_value = []
            for base_value, param_group in zip(self.base_values, self.optimizer.param_groups):
                eta_min = self.eta_min if self.eta_min_ratio is None else base_value * self.eta_min_ratio
                step_ratio = (1 - 1 / (self.num_epochs - self.last_epoch + 1)) ** self.power
                step_value = (param_group['lr'] - eta_min) * step_ratio + eta_min
                param_groups_value.append(step_value)
            return param_groups_value
    # class PolyLRRatio(LambdaLR):
    #     def __init__(self, optimizer, eta_min=0, eta_min_ratio=None, power=1.0, num_epochs=100, warmup_epochs=0,
    #                  last_epoch=-1):
    #         self.eta_min = eta_min
    #         self.eta_min_ratio = eta_min_ratio
    #         self.power = power
    #         self.num_epochs = num_epochs
    #         self.warmup_epochs = warmup_epochs
    #         self.base_values = [group['lr'] for group in optimizer.param_groups]
    #
    #         # Define the lambda function here
    #         self.lr_lambda = self._get_lr_lambda()
    #         super().__init__(optimizer, self.lr_lambda, last_epoch)
    #
    #     def _get_lr_lambda(self):
    #         # Return a function or a list of functions for LambdaLR
    #         def lr_lambda(epoch):
    #             if epoch < self.warmup_epochs:
    #                 return (1.0 / (self.warmup_epochs + 1)) * (epoch + 1)
    #
    #             factor = (1 - (epoch - self.warmup_epochs) / float(self.num_epochs - self.warmup_epochs)) ** self.power
    #             return [(base_lr - self.eta_min) * factor + self.eta_min for base_lr in self.base_values][0]
    #
    #         return lr_lambda
    #
    #     def _get_value(self):
    #         param_groups_value = []
    #         for base_value, param_group in zip(self.base_values, self.optimizer.param_groups):
    #             eta_min = self.eta_min if self.eta_min_ratio is None else base_value * self.eta_min_ratio
    #             step_ratio = (1 - 1 / (self.num_epochs - self.last_epoch + 1)) ** self.power
    #             step_value = (param_group['lr'] - eta_min) * step_ratio + eta_min
    #             param_groups_value.append(step_value)
    #         return param_groups_value
    def adjust_learning_rate(self,optimizer, epoch, params_alr,step_size):
        """Sets the learning rate to the initial LR decayed by 10 every 30 epochs"""
        lr = params_alr["lr"] * (0.1 ** (epoch // step_size))
        for param_group in optimizer.param_groups:
            param_group['lr'] = lr

    def create_lr_scheduler(self,optimizer, num_step, epochs, method="poly", warmup=True, warmup_epochs=1,
                            warmup_factor=1e-3, params_alr=None):
        if method == "step":
            def f(x):
                self.adjust_learning_rate(optimizer, x, params_alr,num_step)
                return optimizer.param_groups[0]['lr'] / params_alr['lr']

            return torch.optim.lr_scheduler.LambdaLR(optimizer, lr_lambda=f)
        elif method == "poly":
            return self.PolyLRRatio(optimizer, eta_min=0, eta_min_ratio=0.1, power=0.9, num_epochs=epochs,
                               warmup_epochs=warmup_epochs)
        else:
            raise ValueError(f"Unknown learning rate method: {method}")


    def accuracy(self,output, target, topk=(1,)):
        """Computes the accuracy over the k top predictions for the specified values of k"""
        with torch.no_grad():
            maxk = max(topk)
            batch_size = target.size(0)

            _, pred = output.topk(maxk, 1, True, True)
            pred = pred.t()
            correct = pred.eq(target.view(1, -1).expand_as(pred))

            res = []
            for k in topk:
                correct_k = correct[:k].reshape(-1).float().sum(0, keepdim=True)
                res.append(correct_k.mul_(100.0 / batch_size))
            return res





import os
import struct
import torch

class gen_wts:
    def __init__(self, weights, output, model_type):
        self.weights = weights
        self.output = output
        self.model_type = model_type

    @classmethod
    def from_args(cls, weights, output=None, model_type='detect'):
        if not os.path.isfile(weights):
            raise ValueError('Invalid input file')
        if not output:
            output = os.path.splitext(weights)[0] + '.wts'
        elif os.path.isdir(output):
            output = os.path.join(output, os.path.splitext(os.path.basename(weights))[0] + '.wts')
        return cls(weights, output, model_type)

    def convert(self):
        print(f'Generating .wts for {self.model_type} model')
        print(f'Loading {self.weights}')

        device = 'cpu'
        # Load model
        #判断模型中是否有model或ema
        themodel = torch.load(self.weights, map_location=device)
        if themodel['model'] is not None:
            thismodel = themodel['model']
        elif themodel['ema'] is not None:
            thismodel = themodel['ema']
        else:
            raise ValueError("Both 'model' and 'ema' are None.")
        # 判断模型中是否有model或ema
        model = thismodel.float()  # load to FP32model = torch.load(pt_file, map_location=device)['model'].float()  # load to FP32
        # model = torch.load(self.weights, map_location=device)['model'].float()

        if self.model_type in ['detect', 'seg']:
            anchor_grid = model.model[-1].anchors * model.model[-1].stride[..., None, None]
            delattr(model.model[-1], 'anchors')

        model.to(device).eval()

        with open(self.output, 'w') as f:
            f.write('{}\n'.format(len(model.state_dict().keys())))
            for k, v in model.state_dict().items():
                vr = v.reshape(-1).cpu().numpy()
                f.write('{} {} '.format(k, len(vr)))
                for vv in vr:
                    f.write(' ')
                    f.write(struct.pack('>f', float(vv)).hex())
                f.write('\n')

# # Example usage
# converter = gen_wts.from_args(
#     weights=r'E:\ALLvision\pycharmproject\Efficientnet\segjjc_clsseg_try\sahi_preprocess\1014\try\segjjc\yoloout\train12\weights\best.pt',
# )
# converter.convert()



import os
import cv2
import torch
import numpy as np
from tqdm import tqdm
import json
from PIL import Image
import io
import base64
import copy
import random
import shutil
import yaml
import sys
from ultralytics import YOLO
from types import SimpleNamespace
from .fcn import modelFCN
from .EfficientNet import modelEfficient
from .othermodules import HSAHI_det,PredictSahi_yolo,gen_wts,ROIT_det,PredictSahi_fcn,export_model

def convert(size, box):
    dw = 1. / (size[0])
    dh = 1. / (size[1])
    x = (box[0] + box[1]) / 2.0 - 1
    y = (box[2] + box[3]) / 2.0 - 1
    w = box[1] - box[0]
    h = box[3] - box[2]
    x = x * dw
    w = w * dw
    y = y * dh
    h = h * dh
    return x, y, w, h

def HdictDir(jsonpath, target_path, imgout_path):
    with open(jsonpath, 'r', encoding='utf-8', errors='ignore') as f:
        content = f.read()
    content = json.loads(content)
    image_dir = content.get("image_dir")
    class_names = content.get("class_names")
    new_paths_image = []
    new_paths_segment = []

    for sample in content.get("samples", []):
        image_file_name = sample.get("image_file_name")
        image_id = sample.get("image_id")

        image_saveid = image_file_name.split('/')[-1].split('.')[0]
        out_file = open(target_path + '/%s.txt' % (image_saveid), 'w')
        image_path = os.path.join(image_dir, image_file_name)

        with Image.open(image_path) as img:
            width, height = img.size
        bbox_label_ids = sample.get("bbox_label_id")
        bbox_rows1 = sample.get("bbox_row1")
        bbox_cols1 = sample.get("bbox_col1")
        bbox_rows2 = sample.get("bbox_row2")
        bbox_cols2 = sample.get("bbox_col2")

        if isinstance(bbox_label_ids, list):
            for i in range(len(bbox_label_ids)):
                bbox_label_id = bbox_label_ids[i]
                bbox_row1 = bbox_rows1[i]
                bbox_col1 = bbox_cols1[i]
                bbox_row2 = bbox_rows2[i]
                bbox_col2 = bbox_cols2[i]

                if bbox_col2 > width:
                    bbox_col2 = width
                if bbox_row2 > height:
                    bbox_row2 = height
                b = (bbox_col1, bbox_col2, bbox_row1, bbox_row2)
                bb = convert((width, height), b)
                out_file.write(str(bbox_label_id) + " " + " ".join([str(a) for a in bb]) + '\n')
        else:
            bbox_label_id = bbox_label_ids
            bbox_row1 = bbox_rows1
            bbox_col1 = bbox_cols1
            bbox_row2 = bbox_rows2
            bbox_col2 = bbox_cols2

            if bbox_col2 > width:
                bbox_col2 = width
            if bbox_row2 > height:
                bbox_row2 = height
            b = (bbox_col1, bbox_col2, bbox_row1, bbox_row2)
            bb = convert((width, height), b)
            out_file.write(str(bbox_label_id) + " " + " ".join([str(a) for a in bb]) + '\n')

        shutil.copy(image_path, imgout_path)
    return class_names

def yamlcreate(class_names, imgdiroupaths, yamlsavepath):
    names = {}
    ##########################优化部分##############################
    ##########################优化部分##############################
    ##########################优化部分##############################
    ##########################优化部分##############################
    ##########################优化部分##############################
    if isinstance(class_names,list):
        for i, class_name in enumerate(class_names):
            names[i] = class_name
            ncs=len(class_names)
    else:
        names[0]=class_names
        ncs = 1
    ##########################优化部分##############################
    ##########################优化部分##############################
    ##########################优化部分##############################
    ##########################优化部分##############################
    ##########################优化部分##############################
    traindirtxt = os.path.join(imgdiroupaths, "train.txt").replace("\\", "/")
    valdirtxt = os.path.join(imgdiroupaths, "val.txt").replace("\\", "/")
    data = {
        "train": traindirtxt,
        "val": valdirtxt,
        "names": names,
        "nc": len(class_names)
    }

    savepath = os.path.join(yamlsavepath, "mydata.yaml")
    with open(savepath, 'w', encoding='utf-8') as file:
        yaml.dump(data, file, default_flow_style=False, allow_unicode=True)
    return savepath

def split_train_val(jsonpath, txtsavepath, train_percent, trainval_percent):
    with open(jsonpath, 'r', encoding='utf-8', errors='ignore') as f:
        content = f.read()
    content = json.loads(content)
    samples = content.get("samples")
    num = len(samples)
    list_index = range(num)
    tv = int(num * trainval_percent)
    tr = int(tv * train_percent)
    trainval = random.sample(list_index, tv)
    train = random.sample(trainval, tr)

    file_trainval = open(txtsavepath + '/trainval.txt', 'w')
    file_test = open(txtsavepath + '/test.txt', 'w')
    file_train = open(txtsavepath + '/train.txt', 'w')
    file_val = open(txtsavepath + '/val.txt', 'w')

    for i in list_index:
        savename = samples[i]["image_file_name"].split('/')[-1]
        name = savename + '\n'
        if i in trainval:
            file_trainval.write(name)
            if i in train:
                file_train.write(name)
            else:
                file_val.write(name)
        else:
            file_test.write(name)

def saveimgdir(txtsavepath, imgdiroupath, imgdir_all):
    sets = ['train', 'val', 'test']
    for image_set in sets:
        image_ids = open(txtsavepath + '/%s.txt' % (image_set)).read().strip().split()
        list_file = open(imgdiroupath + '/%s.txt' % (image_set), 'w')

        for image_id in image_ids:
            list_file.write(imgdir_all + '/%s\n' % (image_id))

def halcon2yolos(params):
    alldir = params['data_alldir']
    jsonpath = params['data_jsonpath']
    if alldir is None:
        script_dir = os.path.dirname(os.path.abspath(sys.argv[0]))
        default_alldir = os.path.join(script_dir, 'Data')
        if not os.path.exists(default_alldir):
            os.makedirs(default_alldir)
        alldir = default_alldir
    imgdiroupath = os.path.join(alldir, 'paper_data')
    txtsavepath = os.path.join(alldir, 'Imagesets')
    target_path = os.path.join(alldir, 'labels')
    imgout_path = os.path.join(alldir, 'images')
    imgdir_all = imgout_path.replace("\\", "/")
    if not os.path.exists(imgdiroupath):
        os.makedirs(imgdiroupath)
    if not os.path.exists(txtsavepath):
        os.makedirs(txtsavepath)
    if not os.path.exists(target_path):
        os.makedirs(target_path)
    if not os.path.exists(imgout_path):
        os.makedirs(imgout_path)
    trainval_percent = 1.0
    train_percent = 0.7
    split_train_val(jsonpath, txtsavepath, train_percent, trainval_percent)
    saveimgdir(txtsavepath, imgdiroupath, imgdir_all)
    class_names = HdictDir(jsonpath, target_path, imgout_path)
    yamldir = yamlcreate(class_names, imgdiroupath, alldir)
    return yamldir

def trainyolo(params,datadir):
    # 加载 default.yaml 文件
    from ultralytics.utils import DEFAULT_CFG,DEFAULT_CFG_DICT
    original_cfg = copy.deepcopy(DEFAULT_CFG)# 深拷贝初始配置
    original_cfg_dict=copy.deepcopy(DEFAULT_CFG_DICT)

    if (params["save_dir"]):
        ####创建训练输出子文件train####
        from ultralytics.utils.files import increment_path
        train_childdir=os.path.join(params["save_dir"],'train')
        train_dir = str(increment_path(train_childdir, exist_ok=False))
        ######
        setattr(DEFAULT_CFG,'save_dir',train_dir)
        DEFAULT_CFG_DICT['save_dir'] = train_dir
#######使用sahi预先对大图进行切图#####
#######使用sahi预先对大图进行切图#####
    if (params.get("sahi_or_roi", None)):
        sahi_or_roi=params["sahi_or_roi"]
        if sahi_or_roi in ['sahi', 'roisahi']:
            sahi=HSAHI_det(yaml_path=datadir,
                           slice_width=params["imgsz"],
                           slice_height=params["imgsz"],
                           overlap_height_ratio=params.get("sahi_overlapratio", 0.25),  # 使用get方法
                           overlap_width_ratio=params.get("sahi_overlapratio", 0.25),
                           sahi_datadir=params.get("sahi_datadir", None),
                           roi=params.get("roi", None))
            sahi.process_data()
            datadir=sahi.output_yaml_path
        if sahi_or_roi=='roi':
            RoiT=ROIT_det(yaml_path=datadir,
                          roi_datadir=params.get("roi_datadir", None),
                          roi=params.get("roi", None))
            RoiT.process_data()
            datadir=RoiT.output_yaml_path
#######使用sahi预先对大图进行切图#####
#######使用sahi预先对大图进行切图#####
    model = YOLO(params["modelpath"])
    model.train(data=datadir,
                epochs=params["epochs"],
                batch=params["batch"],
                workers=params["workers"],
                amp=params["amp"],
                imgsz=params["imgsz"],
                patience=params["patience"],
                device=params["device"],
                resume=params["resume"],
                save_period=params["save_period"],
                optimizer=params["optimizer"],
                seed=params["seed"],
                cos_lr=params["cos_lr"],
                close_mosaic=params["close_mosaic"],
                lr0=params["lr0"],
                lrf=params["lrf"],
                momentum=params["momentum"],
                weight_decay=params["weight_decay"],
                warmup_epochs=params["warmup_epochs"],
                warmup_momentum=params["warmup_momentum"],
                warmup_bias_lr=params["warmup_bias_lr"],
                box=params["box"],
                cls=params["cls"],
                dfl=params["dfl"],
                pose=params["pose"],
                kobj=params["kobj"],
                label_smoothing=params["label_smoothing"],
                nbs=params["nbs"],
                hsv_h=params["hsv_h"],
                hsv_s=params["hsv_s"],
                hsv_v=params["hsv_v"],
                degrees=params["degrees"],
                translate=params["translate"],
                scale=params["scale"],
                shear=params["shear"],
                perspective=params["perspective"],
                flipud=params["flipud"],
                fliplr=params["fliplr"],
                bgr=params["bgr"],
                mosaic=params["mosaic"],
                mixup=params["mixup"],
                copy_paste=params["copy_paste"],
                auto_augment=params["auto_augment"],
                erasing=params["erasing"],
                crop_fraction=params["crop_fraction"])
    ####更改val输出路径#####
    DEFAULT_CFG_DICT['save_dir']=os.path.join(model.trainer.save_dir,'val')
    model.overrides['save_dir']=os.path.join(model.trainer.save_dir,'val')
    ####
    metrics = model.val()
    modelsavedir = str(model.trainer.best)
    weightdir = str(model.trainer.wdir)
    if (params["save_dir"]):
        modeldirs=modelsavedir
    else:
        script_dir = os.path.dirname(os.path.abspath(sys.argv[0]))
        modeldirs = os.path.join(script_dir, modelsavedir)

    # # 将配置还原回原来的样子
    DEFAULT_CFG=original_cfg
    DEFAULT_CFG_DICT=original_cfg_dict
    return modeldirs, weightdir,datadir

def export_yolo(params,model_path):
    model = YOLO(model_path)
    ####wts模型导出，目前默认detect，如果是其他的任务，就需要指定参数model_type=,以及修改gen_wts代码
    if params['export_format']=='wts':
        converter = gen_wts.from_args(
            weights=model_path,
        )
        converter.convert()
    else:
        model.export(format=params['export_format'], imgsz=params['imgsz_export'], half=params['half'], dynamic=params['dynamic'], simplify= params['simplify'])

def export_fcn(params_export,model_path):
    ##此处怎么像yolo一样能够直接根据模型pt重构，而不是来自训练FCNParams的params重新初始化类
    # region 原版导出模型
    # modelfcn_export=modelFCN(params)
    # fcnmodel_toexport = modelfcn_export.export_model(model_path)
    # endregion
    #不过在这里好像主要用类里面的export_model函数，参数是什么不重要，只要能够初始化类用到函数，也可以考虑直接把函数拿出来用
    fcnmodel_toexport=export_model.export_model_fcn(model_path)
    output_onnx_path = model_path.split('.')[0]+'.'+params_export["export_format"]
    # ####输出模型
    # torch.save(fcnmodel_toexport, model_path.split('.')[0]+'xx.pth')
    # ####
    # 2. 定义输入张量
    if params_export["export_format"]=='onnx':
        imgsz_input=params_export["imgsz_export"]
        dummy_input = torch.randn(1, 3, imgsz_input[0],imgsz_input[1])  # 假设输入形状为 (batch_size, channels, height, width)
        torch.onnx.export(fcnmodel_toexport, dummy_input, output_onnx_path, verbose=params_export["export_detail"], opset_version=11)
        print("✅ 模型已导出为 ONNX！\n导出路径:",output_onnx_path)
    if params_export["export_format"] == 'wts':
        import struct
        f = open(output_onnx_path, 'w')
        f.write("{}\n".format(len(fcnmodel_toexport.state_dict().keys())))
        for k,v in fcnmodel_toexport.state_dict().items():
            if params_export["export_detail"]:
                print('key: ', k)
                print('value: ', v.shape)
            vr = v.reshape(-1).cpu().numpy()
            f.write("{} {}".format(k, len(vr)))
            for vv in vr:
                f.write(" ")
                f.write(struct.pack(">f", float(vv)).hex())
            f.write("\n")

def export_efficient(params_export,model_path):
    # region 原版导出写法，要初始化模型，不合适
    # modeleffic_export=modelEfficient(params)
    # efficmodel_toexport=modeleffic_export.export_model(model_path)
    # endregion
    #新版导出模型
    efficmodel_toexport = export_model.export_model_efficient(model_path)
    output_onnx_path = model_path.split('.')[0] + '.' + params_export["export_format"]
    if params_export["export_format"]=='onnx':
        imgsz_input=params_export["imgsz_export"]
        dummy_input = torch.randn(1, 3, imgsz_input[0],imgsz_input[1])  # 假设输入形状为 (batch_size, channels, height, width)
        # 检查模型是否被封装在 DataParallel 中
        if isinstance(efficmodel_toexport, torch.nn.DataParallel):
            # 提取原始模型
            efficmodel_toexport = efficmodel_toexport.module
        elif isinstance(efficmodel_toexport.features, torch.nn.DataParallel):
            # 提取原始模型
            efficmodel_toexport = efficmodel_toexport.features.module
        torch.onnx.export(efficmodel_toexport, dummy_input, output_onnx_path, verbose=params_export["export_detail"], opset_version=11)
    if params_export["export_format"] == 'wts':
        import struct
        f = open(output_onnx_path, 'w')
        f.write("{}\n".format(len(efficmodel_toexport.state_dict().keys())))
        for k,v in efficmodel_toexport.state_dict().items():
            if params_export["export_detail"]:
                print('key: ', k)
                print('value: ', v.shape)
            vr = v.reshape(-1).cpu().numpy()
            f.write("{} {}".format(k, len(vr)))
            for vv in vr:
                f.write(" ")
                f.write(struct.pack(">f", float(vv)).hex())
            f.write("\n")

def testimage_yolo(params,yaml_datadir,infermodeldir,testformat):
    format = testformat
    if format == "onnx":
        infermodel = os.path.join(infermodeldir, "best.onnx")
    elif format == "openvino":
        infermodel = os.path.join(infermodeldir, "best_openvino_model")
    elif format == "engine":
        infermodel = os.path.join(infermodeldir, "best.engine")
    elif format == "pt" or format == "wts":
        infermodel = os.path.join(infermodeldir, "best.pt")

    predictor_test=PredictSahi_yolo(infermodel,params)
    ##判断yaml_datadir是否为None，即是否在测试之前就有训练步骤
    if yaml_datadir:
        predictor_test.predict_based_on_size(yaml_datadir)
    else:
        predictor_test.predict_based_on_size()

def testimage_fcn(params,infermodeldir,testformat,yaml_datadir=None):
    format = testformat
    if format == "onnx":
        infermodel = os.path.join(infermodeldir, "best.onnx")
    elif format == "openvino":
        # infermodel = os.path.join(infermodeldir, "best_openvino_model")
        #目前先用onnx
        infermodel = os.path.join(infermodeldir, "best.onnx")
    elif format == "engine":
        infermodel = os.path.join(infermodeldir, "best.engine")
    elif format == "pt" or format == "wts":
        infermodel = os.path.join(infermodeldir, "best.pt")
    elif format=="pth":
        infermodel = os.path.join(infermodeldir, "best.pth")

    predictor_test=PredictSahi_fcn(infermodel,params)
    ##判断yaml_datadir是否为None，即是否在测试之前就有训练步骤
    if yaml_datadir:
        predictor_test.predict_normal()
    else:
        predictor_test.predict_normal()

def trainfcn(params):
    modelfcn=modelFCN(params)
    modelfcn.train()
    return modelfcn.model_savedir

def trainEfficientNet(params):
    modelefficientNet=modelEfficient(params)
    modelefficientNet.train_Efficient()
    return modelefficientNet.model_savedir
def yoloit(alljsondir,aimode=None):
    with open(alljsondir, 'r',encoding='utf-8', errors='ignore') as f:
        Allparams = json.load(f)
    DefaultParams=Allparams["DefualtParams"]
    MakeDataParams = Allparams["MakeDataParams"]
    TrainParams = Allparams["TrainParams"]
    ExportParams = Allparams["ExportParams"]
    TestParams = Allparams["TestParams"]
    makedata = DefaultParams['ifmkdata']
    iftrain = DefaultParams['iftrain']
    ifexport = DefaultParams['ifexport']
    iftest = DefaultParams['iftest']
    modeltype = DefaultParams['modeltype']
    ##判断用哪个接口
    if aimode=='train':
        DefaultParams["iftrain"]=True
        iftrain=DefaultParams["iftrain"]
    if aimode=='export':
        DefaultParams["ifexport"]=True
        ifexport = DefaultParams["ifexport"]
    if aimode=='test':
        DefaultParams["iftest"]=True
        iftest = DefaultParams["iftest"]
    ###是否制作数据集
    if makedata:
        datadir = halcon2yolos(MakeDataParams)
        # datadir = halcon2yolos(jsonpath, alldir=alldir)
    else:
        datadir=None
    ###是否进行训练
    if iftrain:
        if modeltype=='yolo':
            YOLOParams=TrainParams['YOLOParams']
            modeldir, weightdir,yaml_datadir = trainyolo(YOLOParams,datadir=datadir if datadir else YOLOParams["datadir"])
            # modeldir, weightdir = trainyolo(datadir=datadir,
            #                                 modelpath=modelpath,
            #                                 epochs=epochs, batch=batch,
            #                                 workers=workers,
            #                                 amp=amp,
            #                                 imgsz=imgsz,
            #                                 patience=patience,
            #                                 device=device,
            #                                 resume=resume)
        if modeltype == 'fcn':
            FCNParams = TrainParams['FCNParams']
            modeldir=trainfcn(FCNParams)

        if modeltype == 'efficient':
            EfficientNetParams = TrainParams['EfficientNetParams']
            modeldir=trainEfficientNet(EfficientNetParams)
    else:
        modeldir, weightdir,yaml_datadir=None,None,None
    ###是否导出模型
    if ifexport:
        if modeltype == 'yolo':
            ExportParams=ExportParams["ExportParams_YOLO"]
            export_yolo(ExportParams,model_path=modeldir if modeldir else ExportParams["model_to_export"])
            weightdir_export=weightdir if weightdir else os.path.dirname(ExportParams["model_to_export"])
            exportmodel_format=ExportParams["export_format"]
        if modeltype == 'fcn':
            ExportParams = ExportParams["ExportParams_FCN"]
            # export_fcn(FCNParams, ExportParams, model_path=modeldir)
            export_fcn(ExportParams,model_path=modeldir if modeldir else ExportParams["model_to_export"])
            weightdir_export = os.path.dirname(modeldir) if modeldir else os.path.dirname(ExportParams["model_to_export"])
            exportmodel_format = ExportParams["export_format"]
            #######继续修改，怎么导出########
        if modeltype == 'efficient':
            ExportParams = ExportParams["ExportParams_EfficientNet"]
            # export_efficient(EfficientNetParams, ExportParams, model_path=modeldir)
            export_efficient(ExportParams,model_path=modeldir if modeldir else ExportParams["model_to_export"])#######继续修改，怎么导出########
    else:
        weightdir_export=None
        exportmodel_format = None
    ###是否进行测试
    if iftest:
        if modeltype == 'yolo':
            TestParams = TestParams["TestParams_YOLO"]
            # testimage(cur_path=inferedimg, infermodeldir=weightdir_export if weightdir_export else params["infermodeldir"],
            #           testimgpath=testimg, format=format)
            ##如果训练时输出test图片的路径就放在训练输出一起，训练参数还需要添加一个“da_testimg”:true/false,如果json文件没有这个关键词，默认为True
            if weightdir and TrainParams.get("do_testimg",True):
                test_oudir=os.path.join(os.path.dirname(weightdir),'test')
                if not os.path.exists(test_oudir):
                    os.makedirs(test_oudir)
                autolabel_dir = os.path.join(test_oudir, 'autolabel')
                inferedimg_dir=os.path.join(test_oudir, 'inferedimg')
                if not os.path.exists(autolabel_dir):
                    os.makedirs(autolabel_dir)
                if not os.path.exists(inferedimg_dir):
                    os.makedirs(inferedimg_dir)
                TestParams["resultjson_path"]=autolabel_dir
                TestParams["inferedimg"] = inferedimg_dir
            ##如果训练时输出test图片的路径就放在训练输出一起
            testimage_yolo(TestParams,yaml_datadir,infermodeldir=weightdir_export if weightdir_export else TestParams["infermodeldir"],testformat=exportmodel_format if exportmodel_format else TestParams["infer_format"])

        if modeltype == 'fcn':
            TestParams = TestParams["TestParams_FCN"]
            testimage_fcn(TestParams,
                          infermodeldir=weightdir_export if weightdir_export else TestParams["infermodeldir"],
                          testformat=exportmodel_format if exportmodel_format else TestParams["infer_format"])
        if modeltype == 'efficient':
            print('efficient test not support now!! to developing')

def yoloit_solo(alljsondir,aimode=None):
    with open(alljsondir, 'r',encoding='utf-8', errors='ignore') as f:
        Allparams = json.load(f)
    DefaultParams=Allparams["DefualtParams"]
    MakeDataParams = Allparams["MakeDataParams"]
    TrainParams = Allparams["TrainParams"]
    ExportParams = Allparams["ExportParams"]
    TestParams = Allparams["TestParams"]
    makedata = DefaultParams.get('ifmkdata', False)
    iftrain = DefaultParams.get('iftrain', False)
    ifexport = DefaultParams.get('ifexport', False)
    iftest = DefaultParams.get('iftest', False)
    modeltype = DefaultParams['modeltype']
    ##判断用哪个接口
    if aimode=='train':
        DefaultParams["ifmkdata"]=False
        ifmkdata=DefaultParams["ifmkdata"]
        DefaultParams["iftrain"]=True
        iftrain=DefaultParams["iftrain"]
        DefaultParams["ifexport"]=False
        ifexport=DefaultParams["ifexport"]
        DefaultParams["iftest"]=False
        iftest=DefaultParams["iftest"]
    if aimode=='export':
        DefaultParams["ifmkdata"]=False
        ifmkdata=DefaultParams["ifmkdata"]
        DefaultParams["iftrain"]=False
        iftrain=DefaultParams["iftrain"]
        DefaultParams["ifexport"]=True
        ifexport=DefaultParams["ifexport"]
        DefaultParams["iftest"]=False
        iftest=DefaultParams["iftest"]
    if aimode=='test':
        DefaultParams["ifmkdata"]=False
        ifmkdata=DefaultParams["ifmkdata"]
        DefaultParams["iftrain"]=False
        iftrain=DefaultParams["iftrain"]
        DefaultParams["ifexport"]=False
        ifexport=DefaultParams["ifexport"]
        DefaultParams["iftest"]=True
        iftest=DefaultParams["iftest"]
    ###是否制作数据集
    if makedata:
        datadir = halcon2yolos(MakeDataParams)
        # datadir = halcon2yolos(jsonpath, alldir=alldir)
    else:
        datadir=None
    ###是否进行训练
    if iftrain:
        if modeltype=='yolo':
            YOLOParams=TrainParams['YOLOParams']
            modeldir, weightdir,yaml_datadir = trainyolo(YOLOParams,datadir=datadir if datadir else YOLOParams["datadir"])
            # modeldir, weightdir = trainyolo(datadir=datadir,
            #                                 modelpath=modelpath,
            #                                 epochs=epochs, batch=batch,
            #                                 workers=workers,
            #                                 amp=amp,
            #                                 imgsz=imgsz,
            #                                 patience=patience,
            #                                 device=device,
            #                                 resume=resume)
        if modeltype == 'fcn':
            FCNParams = TrainParams['FCNParams']
            modeldir=trainfcn(FCNParams)

        if modeltype == 'efficient':
            EfficientNetParams = TrainParams['EfficientNetParams']
            modeldir=trainEfficientNet(EfficientNetParams)
    else:
        modeldir, weightdir,yaml_datadir=None,None,None
    ###是否导出模型
    if ifexport:
        if modeltype == 'yolo':
            ExportParams=ExportParams["ExportParams_YOLO"]
            export_yolo(ExportParams,model_path=modeldir if modeldir else ExportParams["model_to_export"])
            weightdir_export=weightdir if weightdir else os.path.dirname(ExportParams["model_to_export"])
            exportmodel_format=ExportParams["export_format"]
        if modeltype == 'fcn':
            ExportParams = ExportParams["ExportParams_FCN"]
            export_fcn(FCNParams,ExportParams,model_path=modeldir)#######继续修改，怎么导出########
        if modeltype == 'efficient':
            ExportParams = ExportParams["ExportParams_EfficientNet"]
            export_efficient(EfficientNetParams,ExportParams,model_path=modeldir)#######继续修改，怎么导出########
    else:
        weightdir_export=None
        exportmodel_format = None
    ###是否进行测试
    if iftest:
        if modeltype == 'yolo':
            TestParams = TestParams["TestParams_YOLO"]
            # testimage(cur_path=inferedimg, infermodeldir=weightdir_export if weightdir_export else params["infermodeldir"],
            #           testimgpath=testimg, format=format)
            ##如果训练时输出test图片的路径就放在训练输出一起，训练参数还需要添加一个“da_testimg”:true/false,如果json文件没有这个关键词，默认为True
            if weightdir and TrainParams.get("do_testimg",True):
                test_oudir=os.path.join(os.path.dirname(weightdir),'test')
                if not os.path.exists(test_oudir):
                    os.makedirs(test_oudir)
                if TestParams.get("resultjson_path",False):
                    autolabel_dir = os.path.join(test_oudir, 'autolabel')
                    if not os.path.exists(autolabel_dir):
                        os.makedirs(autolabel_dir)
                inferedimg_dir = os.path.join(test_oudir, 'inferedimg')
                if not os.path.exists(inferedimg_dir):
                    os.makedirs(inferedimg_dir)
                TestParams["resultjson_path"]=autolabel_dir
                TestParams["inferedimg"] = inferedimg_dir
            ##如果训练时输出test图片的路径就放在训练输出一起
            testimage_yolo(TestParams,yaml_datadir,infermodeldir=weightdir_export if weightdir_export else TestParams["infermodeldir"],testformat=exportmodel_format if exportmodel_format else TestParams["infer_format"])

        if modeltype == 'fcn':
            TestParams = TestParams["TestParams_FCN"]

        if modeltype == 'efficient':
            print('efficient test not support now!! to developing')



import numpy as np
import random
from typing import Tuple

import torch
from torchvision import transforms as T
from torchvision.transforms import functional as F
from PIL import Image


def letterbox_pil(image: Image.Image, new_shape: Tuple[int, int] = (640, 640), auto=True, scaleFill=False, scaleup=True,
                  stride=32):
    """
    Resize and pad PIL Image while meeting stride-multiple constraints.

    Args:
        image (Image.Image): The input PIL Image to be resized and padded.
        new_shape (Tuple[int, int]): The desired output shape of the image.
        auto (bool): Whether to keep the original ratio and pad the image.
        scaleFill (bool): Whether to stretch the image to fit the new shape.
        scaleup (bool): Whether to scale up the image.
        stride (int): Stride used for padding.

    Returns:
        Image.Image: Resized and padded PIL Image.
    """
    # Compute scale ratio and new shape
    shape = image.size  # current shape [width, height]
    if isinstance(new_shape, int):
        new_shape = (new_shape, new_shape)

    # Scale ratio (new / old)
    r = min(new_shape[0] / shape[0], new_shape[1] / shape[1])
    if not scaleup:  # only scale down, do not scale up (for better val mAP)
        r = min(r, 1.0)

    # Compute padding
    ratio = r, r  # width, height ratios
    new_unpad = int(round(shape[0] * r)), int(round(shape[1] * r))
    dw, dh = new_shape[0] - new_unpad[0], new_shape[1] - new_unpad[1]  # wh padding
    if auto:  # minimum rectangle
        dw, dh = np.mod(dw, stride), np.mod(dh, stride)  # wh padding
    elif scaleFill:  # stretch
        dw, dh = 0.0, 0.0
        new_unpad = (new_shape[0], new_shape[1])
        ratio = new_shape[0] / shape[1], new_shape[1] / shape[0]  # width, height ratios

    dw /= 2  # divide padding into 2 sides
    dh /= 2

    # Resize the image
    mode = 'L' if image.mode == 'L' else 'RGB'
    resized_image = image.resize(new_unpad, Image.BILINEAR).convert(mode)

    # Create a new image with the desired background color
    if mode == 'L':
        new_image = Image.new('L', new_shape, 0)  # 单通道图像填充为 114
    else:
        new_image = Image.new('RGB', new_shape, (114, 114, 114))  # 三通道图像填充为 (114, 114, 114)

    # Calculate the position to paste the resized image
    paste_left = int(round(dw))
    paste_top = int(round(dh))
    paste_right = int(round(paste_left + new_unpad[0]))
    paste_bottom = int(round(paste_top + new_unpad[1]))

    # Paste the resized image onto the new image
    new_image.paste(resized_image, (paste_left, paste_top, paste_right, paste_bottom))

    return new_image

def pad_if_smaller(img, size, fill=0):
    # 如果图像最小边长小于给定size，则用数值fill进行padding
    min_size = min(img.size)
    if min_size < size:
        ow, oh = img.size
        padh = size - oh if oh < size else 0
        padw = size - ow if ow < size else 0
        img = F.pad(img, (0, 0, padw, padh), fill=fill)
    return img


class Compose(object):
    def __init__(self, transforms):
        self.transforms = transforms

    def __call__(self, image, target):
        for t in self.transforms:
            image, target = t(image, target)
        return image, target


class RandomResize(object):
    def __init__(self, min_size, max_size=None):
        self.min_size = min_size
        if max_size is None:
            max_size = min_size
        self.max_size = max_size

    def __call__(self, image, target):
        size = random.randint(self.min_size, self.max_size)
        # 这里size传入的是int类型，所以是将图像的最小边长缩放到size大小
        image = F.resize(image, size)
        # 这里的interpolation注意下，在torchvision(0.9.0)以后才有InterpolationMode.NEAREST
        # 如果是之前的版本需要使用PIL.Image.NEAREST
        target = F.resize(target, size, interpolation=T.InterpolationMode.NEAREST)
        return image, target

class imgResize(object):
    def __init__(self,size:Tuple[int, int]):
        self.max_size = (max(size),max(size))

    def __call__(self, image, target):
        image=letterbox_pil(image,self.max_size)
        target = letterbox_pil(target, self.max_size)
        return image, target

class RandomHorizontalFlip(object):
    def __init__(self, flip_prob):
        self.flip_prob = flip_prob

    def __call__(self, image, target):
        if random.random() < self.flip_prob:
            image = F.hflip(image)
            target = F.hflip(target)
        return image, target


class RandomCrop(object):
    def __init__(self, size):
        self.size = size

    def __call__(self, image, target):
        image = pad_if_smaller(image, self.size)
        target = pad_if_smaller(target, self.size, fill=255)
        crop_params = T.RandomCrop.get_params(image, (self.size, self.size))
        image = F.crop(image, *crop_params)
        target = F.crop(target, *crop_params)
        return image, target


class CenterCrop(object):
    def __init__(self, size):
        self.size = size

    def __call__(self, image, target):
        image = F.center_crop(image, self.size)
        target = F.center_crop(target, self.size)
        return image, target


class ToTensor(object):
    def __call__(self, image, target):
        image = F.to_tensor(image)
        target = torch.as_tensor(np.array(target), dtype=torch.int64)
        return image, target


class Normalize(object):
    def __init__(self, mean, std):
        self.mean = mean
        self.std = std

    def __call__(self, image, target):
        image = F.normalize(image, mean=self.mean, std=self.std)
        return image, target


# # 加载图像
# image_path = 'E:/ALLvision/pycharmproject/yoloseg/githubtry/diyFCN/try2_yaml/fcn/Data_seg/images/train/44.jpg'
# target_path = 'E:/ALLvision/pycharmproject/yoloseg/githubtry/diyFCN/try2_yaml/fcn/tryimgsave/1.png'
#
# img = Image.open(image_path)
# target = Image.open(target_path)
#
# # 创建 imgResize 实例
# img_resizer = imgResize(size=(640, 640))
#
# # 调用 imgResize 实例来调整图像大小
# resized_img, resized_target = img_resizer(img, target)
#
# # 显示调整大小后的图像
# resized_img.show()
# resized_target.show()


import os
import cv2
from ultralytics import YOLO
import time
import json
import torch

# # 设置保存图片的路径
# cur_path = r"E:\ALLvision\pycharmproject\pvk\autolabel\try\out\imgsave"
# # 训练好的模型权重路径
# model = YOLO(r"E:\ALLvision\pycharmproject\pvk\autolabel\try\model\best.pt")
# # 测试图片的路径
# file_pathname = r"E:\工作\钙钛矿玻璃\采图\2419LPL\Sourceimg_resizejpg\PL-蓝光\2024_11_15\Source"
# #json输出路径
# outjson_path=r"E:\ALLvision\pycharmproject\pvk\autolabel\try\out\datajson"
# # 存储推理结果的列表
# predictions = []
# file_pathnames=file_pathname+"\\"

class YOLOauto_label_det:
    def __init__(self,params,model,outjson_path=None,labelimg_path=None,labeled_imgdir=None):
        # 判断传入的 model 是 YOLO 实例还是文件路径
        if isinstance(model, YOLO):
            self.model = model  # 如果是 YOLO 实例，直接赋值
        elif isinstance(model, str) and os.path.isfile(model):  # 如果是路径并且文件存在
            self.model = YOLO(model)  # 加载路径对应的模型
        else:
            raise ValueError("Invalid model input. Must be a YOLO instance or a valid file path.")
        self.params=params
        self.outjson_path=params.get("outjson_path", None) if params.get("outjson_path", None) else outjson_path
        self.labelimg_path=params.get("labelimg_path", None) if params.get("labelimg_path", None) else labelimg_path
        self.file_pathnames=self.labelimg_path+"\\"
        self.labeled_imgdir=params.get("labeled_imgdir", None) if params.get("labeled_imgdir", None) else labeled_imgdir
    def extract_predictions(self,result, image_id, image_file_name):
        sample = []
        for det in result:
            # 将张量转换为列表
            xyxy_list = det.boxes.xyxy.tolist()
            # 初始化坐标列表
            col1_list = []
            row1_list = []
            col2_list = []
            row2_list = []
            # 遍历边界框坐标列表
            for box in xyxy_list:
                col1, row1, col2, row2 = box
                col1_list.append(col1)
                row1_list.append(row1)
                col2_list.append(col2)
                row2_list.append(row2)
            sample.append({
                "image_id": image_id,
                "image_file_name": image_file_name,
                "bbox_label_id": det.boxes.cls.to(torch.int).tolist(),
                "bbox_row1": row1_list,
                "bbox_col1": col1_list,
                "bbox_row2": row2_list,
                "bbox_col2": col2_list
            })
        return sample

    def labelit(self):
        # 存储推理结果的列表
        predictions = []
        # 存储JSON结果的列表
        json_results = {
            "class_ids": list(self.model.names.keys()),
            "class_names": list(self.model.names.values()),
            "image_dir": self.file_pathnames.replace("\\", "/"),
            "samples": []
        }
        # 处理文件夹中的每张图片
        for image_id, filename in enumerate(os.listdir(self.labelimg_path)):
            if filename.split('.')[-1] != 'txt':
                img = cv2.imread(os.path.join(self.labelimg_path, filename))
                # 进行推理
                result = self.model.predict(img, batch=self.params.get("batchsize",1), imgsz=1024)
                # 将推理结果存储到列表中
                predictions.append(result)
                # 提取并存储到JSON结果
                json_results["samples"].extend(self.extract_predictions(result, image_id + 1, filename))
                # t=json_results


        json_path = os.path.join(self.outjson_path, 'predictions.json')
        with open(json_path, 'w', encoding='utf-8') as f:
            json.dump(json_results, f, ensure_ascii=False, indent=4)

        if self.labeled_imgdir:
            self.show_result(predictions)
    def show_result(self,predictions):
        # 创建窗口
        cv2.namedWindow("yolo", 0)
        cv2.resizeWindow("yolo", 640, 640)

        exit_flag = False  # 标志变量
        key = cv2.waitKey(0)
        for j, result in enumerate(predictions):
            if exit_flag:  # 检查是否应该退出外部循环
                break

            if isinstance(result, list):
                for res in result:
                    ann = res.plot()
                    if key == ord('q') and j != len(predictions) - 1:
                        cv2.imshow("yolo", ann)
                        key = cv2.waitKey(0)
                    elif key == ord('q') and j == len(predictions) - 1:
                        exit_flag = True  # 设置标志变量为 True
                        break
                    elif key == 27:
                        exit_flag = True  # 设置标志变量为 True
                        break
                    if not os.path.exists(self.labeled_imgdir):
                        os.mkdir(self.labeled_imgdir)
                    # cv2.imwrite(os.path.join(cur_path, f"out{j}.jpg"), ann)

            else:
                ann = result.plot()
                if key == ord('q') and j != len(predictions) - 1:
                    cv2.imshow("yolo", ann)
                    key = cv2.waitKey(0)
                elif key == ord('q') and j == len(predictions) - 1:
                    exit_flag = True  # 设置标志变量为 True
                    break
                elif key == 27:
                    exit_flag = True  # 设置标志变量为 True
                    break

                if not os.path.exists(self.labeled_imgdir):
                    os.mkdir(self.labeled_imgdir)
                cv2.imwrite(os.path.join(self.labeled_imgdir, f"out{j}.jpg"), ann)

        cv2.destroyAllWindows()


# ###使用
# # 设置保存图片的路径
# cur_path_s = r"E:\ALLvision\pycharmproject\Efficientnet\segjjc_clsseg_try\sahi_preprocess\1125\out\imgsave"
# # 训练好的模型权重路径
# model_s = r"E:\ALLvision\pycharmproject\pvk\autolabel\try\model\best.pt"
# # 测试图片的路径
# file_pathname_s = r"E:\工作\钙钛矿玻璃\采图\2419LPL\Sourceimg_resizejpg\PL-蓝光\2024_11_15\Source"
# #json输出路径
# outjson_path_s=r"E:\ALLvision\pycharmproject\Efficientnet\segjjc_clsseg_try\sahi_preprocess\1125\out\datajson"
# ##定义auto_label
# lit=auto_label(model_s,outjson_path_s,file_pathname_s,cur_path_s)
# #执行
# lit.labelit()


import torch
from torch import nn
import SegJJC.fcn.train_utils.distributed_utils as utils
import torchvision.models.segmentation as models

def criterion(inputs, target):
    losses = {}
    for name, x in inputs.items():
        # 忽略target中值为255的像素，255的像素是目标边缘或者padding填充
        losses[name] = nn.functional.cross_entropy(x, target, ignore_index=255)

    if len(losses) == 1:
        return losses['out']

    return losses['out'] + 0.5 * losses['aux']


def evaluate(model, data_loader, device, num_classes):
    model.eval()
    confmat = utils.ConfusionMatrix(num_classes)
    metric_logger = utils.MetricLogger(delimiter="  ")
    header = 'Test:'
    with torch.no_grad():
        for image, target in metric_logger.log_every(data_loader, 100, header):
            image, target = image.to(device), target.to(device)
            output = model(image)
            output = output['out']

            confmat.update(target.flatten(), output.argmax(1).flatten())

        confmat.reduce_from_all_processes()

    # 计算 mean IoU(nobackground)
    _, _, iu = confmat.compute()
    mean_iou_nobackground = iu[1:].mean().item() * 100
    return confmat,mean_iou_nobackground

#deeplabv3的aspppooling归一化层删除
def disable_bn_in_aspp(model):
    # 直接访问并禁用特定的 BatchNorm2d 层
    aspp_pooling_bn = model.classifier[0].convs[4][2]
    aspp_pooling_bn.training = False

def train_one_epoch(model, optimizer, data_loader, device, epoch, lr_scheduler, print_freq=10, scaler=None):
    model.train()
    # 直接访问并禁用特定的 BatchNorm2d 层
    if isinstance(model,models.DeepLabV3):
        disable_bn_in_aspp(model)
    metric_logger = utils.MetricLogger(delimiter="  ")
    metric_logger.add_meter('lr', utils.SmoothedValue(window_size=1, fmt='{value:.6f}'))
    header = 'Epoch: [{}]'.format(epoch)

    for image, target in metric_logger.log_every(data_loader, print_freq, header):
        image, target = image.to(device), target.to(device)
        # 获取 unique 的标签值
        unique_labels, counts = torch.unique(target, return_counts=True)
        with torch.cuda.amp.autocast(enabled=scaler is not None):
            output = model(image)
            loss = criterion(output, target)

        optimizer.zero_grad()
        if scaler is not None:
            scaler.scale(loss).backward()
            scaler.step(optimizer)
            scaler.update()
        else:
            loss.backward()
            optimizer.step()

        lr_scheduler.step()

        lr = optimizer.param_groups[0]["lr"]
        metric_logger.update(loss=loss.item(), lr=lr)

    return metric_logger.meters["loss"].global_avg, lr


def create_lr_scheduler(optimizer,
                        num_step: int,
                        epochs: int,
                        warmup=True,
                        warmup_epochs=1,
                        warmup_factor=1e-3):
    assert num_step > 0 and epochs > 0
    if warmup is False:
        warmup_epochs = 0

    def f(x):
        """
        根据step数返回一个学习率倍率因子，
        注意在训练开始之前，pytorch会提前调用一次lr_scheduler.step()方法
        """
        if warmup is True and x <= (warmup_epochs * num_step):
            alpha = float(x) / (warmup_epochs * num_step)
            # warmup过程中lr倍率因子从warmup_factor -> 1
            return warmup_factor * (1 - alpha) + alpha
        else:
            # warmup后lr倍率因子从1 -> 0
            # 参考deeplab_v2: Learning rate policy
            return (1 - (x - warmup_epochs * num_step) / ((epochs - warmup_epochs) * num_step)) ** 0.9

    return torch.optim.lr_scheduler.LambdaLR(optimizer, lr_lambda=f)



import tkinter as tk
from tkinter import filedialog, messagebox
from enlarge_label import enlarge_label_det
import sys

class InitialDialog:
    def __init__(self, root):
        self.root = root
        self.root.title("是否已有 Halcon")
        self.root.geometry("300x150")  # 窗口大小

        self.import_HALCONdll = tk.BooleanVar(value=True)  # 默认选择为True（表示安装了Halcon）

        # 提示标签
        tk.Label(root, text="本地是否已经安装 Halcon?").pack(pady=20)

        # 创建选择按钮
        self.install_button = tk.Button(root, text="已安装", command=self.select_install)
        self.install_button.pack(pady=5)

        self.no_install_button = tk.Button(root, text="未安装", command=self.select_no_install)
        self.no_install_button.pack(pady=5)

        # 绑定窗口关闭事件
        self.root.protocol("WM_DELETE_WINDOW", self.forceclose_dialog)

    def select_install(self):
        self.import_HALCONdll.set(False)  # 用户选择安装 Halcon
        self.close_dialog()  # 关闭当前窗口并进入主界面

    def select_no_install(self):
        self.import_HALCONdll.set(True)  # 用户选择不安装 Halcon
        self.close_dialog()  # 关闭当前窗口并进入主界面

    def close_dialog(self):
        """仅关闭选择窗口，不退出整个程序"""
        self.root.quit()  # 退出主循环，关闭当前界面
        self.root.destroy()  # 销毁界面
    def forceclose_dialog(self):
        """仅关闭选择窗口，不退出整个程序"""
        self.root.quit()  # 退出主循环，关闭当前界面
        self.root.destroy()  # 销毁界面
        # 如果你不希望窗口退出，可以调用 sys.exit(0) 来退出程序
        sys.exit(0)  # 完全退出程序

    def get_selection(self):
        return self.import_HALCONdll.get()  # 获取选择的值


# 自定义类用于重定向 print 输出到 Text 控件
class PrintRedirector:
    def __init__(self, text_widget):
        self.text_widget = text_widget

    def write(self, message):
        """将标准输出的消息插入到 Text 控件中"""
        if message != '\n':  # 防止插入空行
            self.text_widget.insert(tk.END, message)
            self.text_widget.see(tk.END)  # 滚动到最新的输出

    def flush(self):
        """flush 函数通常在重定向时需要，但这里我们不需要执行任何操作"""
        pass


# GUI 界面部分
class LabelEnlargerApp:
    def __init__(self, root, import_HALCONdll):
        self.root = root
        self.root.title("标签扩大工具")
        self.root.geometry("600x450")  # 设置初始大小
        self.root.resizable(True, True)  # 允许用户调整窗口大小

        self.model_mode = tk.StringVar()
        self.enlarge_mode = tk.StringVar()
        self.labels = tk.StringVar()
        self.areamins = tk.StringVar()
        self.dict_path = ""
        self.enlarged_dict_path = ""
        self.import_HALCONdll = import_HALCONdll  # 根据选择传入的值

        # 路径选择框 (路径编辑框在前)
        tk.Label(root, text="选择 dict 文件路径").grid(row=2, column=0, padx=10, pady=10, sticky="w")
        self.dict_path_entry = tk.Entry(root, width=40)
        self.dict_path_entry.grid(row=2, column=1, padx=10, pady=10, sticky="ew")
        self.dict_button = tk.Button(root, text="选择文件", command=self.select_dict_path)
        self.dict_button.grid(row=2, column=2, padx=10, pady=10)

        tk.Label(root, text="选择 enlarged_dict 输出路径").grid(row=3, column=0, padx=10, pady=10, sticky="w")
        self.enlarged_dict_entry = tk.Entry(root, width=40)
        self.enlarged_dict_entry.grid(row=3, column=1, padx=10, pady=10, sticky="ew")
        self.enlarged_button = tk.Button(root, text="选择输出", command=self.select_enlarged_dict_path)
        self.enlarged_button.grid(row=3, column=2, padx=10, pady=10)

        # 模式选择框
        tk.Label(root, text="选择模型模式").grid(row=0, column=0, padx=10, pady=10, sticky="w")
        self.model_mode.set("detect")
        tk.OptionMenu(root, self.model_mode, "detect").grid(row=0, column=1, padx=10, pady=10, sticky="ew")

        tk.Label(root, text="选择扩大模式").grid(row=1, column=0, padx=10, pady=10, sticky="w")
        self.enlarge_mode.set("area")
        tk.OptionMenu(root, self.enlarge_mode, "area").grid(row=1, column=1, padx=10, pady=10, sticky="ew")

        # 标签输入框，设置淡颜色字体
        tk.Label(root, text="输入标签 (逗号分隔)").grid(row=4, column=0, padx=10, pady=10, sticky="w")
        label_entry = tk.Entry(root, textvariable=self.labels, fg="lightgray")  # 淡灰色字体
        label_entry.grid(row=4, column=1, padx=10, pady=10, sticky="ew")
        label_entry.insert(0, '晕斑,划痕')

        # 监听标签输入框内容变化
        label_entry.bind("<FocusIn>", lambda event: self.on_focus_in(event, label_entry, '例如: 晕斑,划痕'))
        label_entry.bind("<FocusOut>", lambda event: self.on_focus_out(event, label_entry, '例如: 晕斑,划痕'))

        # 最小面积输入框，设置淡颜色字体
        tk.Label(root, text="输入最小面积 (逗号分隔)").grid(row=5, column=0, padx=10, pady=10, sticky="w")
        area_entry = tk.Entry(root, textvariable=self.areamins, fg="lightgray")  # 淡灰色字体
        area_entry.grid(row=5, column=1, padx=10, pady=10, sticky="ew")
        area_entry.insert(0, '100,100')

        # 监听最小面积输入框内容变化
        area_entry.bind("<FocusIn>", lambda event: self.on_focus_in(event, area_entry, '例如: 100,100'))
        area_entry.bind("<FocusOut>", lambda event: self.on_focus_out(event, area_entry, '例如: 100,100'))

        # 执行按钮
        self.run_button = tk.Button(root, text="执行", command=self.run_enlargement)
        self.run_button.grid(row=6, column=0, columnspan=3, padx=10, pady=10)

        # 控制台输出区域，增大控制台的高度
        self.output_console = tk.Text(root, height=15, width=70)  # 增大控制台区域
        self.output_console.grid(row=7, column=0, columnspan=3, padx=10, pady=10)
        self.output_console.config(state=tk.NORMAL)  # 允许写入

        self.root.grid_rowconfigure(7, weight=1)

        # 将标准输出和错误输出重定向到控制台区域
        sys.stdout = PrintRedirector(self.output_console)
        sys.stderr = PrintRedirector(self.output_console)

    def select_dict_path(self):
        self.dict_path = filedialog.askopenfilename(filetypes=[("HALCON Dict Files", "*.hdict")])
        self.dict_path_entry.delete(0, tk.END)  # 清除旧路径
        self.dict_path_entry.insert(0, self.dict_path)

    def select_enlarged_dict_path(self):
        self.enlarged_dict_path = filedialog.asksaveasfilename(defaultextension=".hdict",
                                                               filetypes=[("HALCON Dict Files", "*.hdict")])
        self.enlarged_dict_entry.delete(0, tk.END)  # 清除旧路径
        self.enlarged_dict_entry.insert(0, self.enlarged_dict_path)

    def on_focus_in(self, event, entry_widget, default_text):
        if entry_widget.get() == default_text:
            entry_widget.delete(0, tk.END)
            entry_widget.config(fg="black")

    def on_focus_out(self, event, entry_widget, default_text):
        if entry_widget.get() == "":
            entry_widget.insert(0, default_text)
            entry_widget.config(fg="lightgray")

    def run_enlargement(self):
        missing_fields = []
        if not self.dict_path:
            missing_fields.append("dict 文件路径")
        if not self.enlarged_dict_path:
            missing_fields.append("enlarged_dict 输出路径")
        if not self.labels.get():
            missing_fields.append("标签")
        if not self.areamins.get():
            missing_fields.append("最小面积")

        if missing_fields:
            print(f"请确保以下字段已填写:\n" + "\n".join(missing_fields))
            return

        labels = self.labels.get().split(",")
        areamins = self.areamins.get().split(",")
        if len(labels) != len(areamins):
            print("标签和最小面积数量不匹配")
            return

        enlargeit = enlarge_label_det(self.dict_path, self.import_HALCONdll)
        enlargeit.enlarge_label_area(labels, [int(x) for x in areamins], self.enlarged_dict_path)

        print(f"标签扩展已完成！")
        print(f"标签: {labels}")
        print(f"最小面积: {areamins}")
        print(f"输出路径: {self.enlarged_dict_path}")
        print("操作成功！")


# 创建程序启动界面，选择是否安装 Halcon
def start_app():
    root = tk.Tk()
    dialog = InitialDialog(root)
    root.mainloop()  # 显示选择界面

    # 获取选择结果，进入主应用
    import_HALCONdll = dialog.get_selection()

    # 启动主应用窗口
    root = tk.Tk()
    app = LabelEnlargerApp(root, import_HALCONdll)
    root.mainloop()


if __name__ == "__main__":
    start_app()



import os
import time
import datetime
import  yaml
import torch
import shutil

from SegJJC.fcn.src import fcn_resnet50,fcn_resnet18,fcn_resnet34,deeplabv3_resnet18,deeplabv3_resnet34#新增resnet18/34
from .train_utils import train_one_epoch, evaluate, create_lr_scheduler
from .my_dataset import myDataset_yaml
import SegJJC.fcn.transforms as T
import torchvision.models.segmentation as models

class modelFCN:
    def __init__(self, params):
        self.params = params
        ###读取train、val的images路径
        with open(self.params["data_yaml_dir"], 'r', encoding='utf-8') as file:
            datayaml = yaml.safe_load(file)
        self.num_classes = datayaml['nc'] + 1
        # 提取 train 和 val 路径
        self.train_path = datayaml['train']
        self.val_path = datayaml['val']
        self.aux=self.params["aux"]
        self.model_savedir=None
        self.best_miou_allcls=0

    class SegmentationPresetTrain:
        def __init__(self, base_size, crop_size, hflip_prob=0.5, mean=(0.485, 0.456, 0.406), std=(0.229, 0.224, 0.225)):
            min_size = int(0.5 * base_size)
            max_size = int(2.0 * base_size)

            trans = [T.RandomResize(min_size, max_size)]
            if hflip_prob > 0:
                trans.append(T.RandomHorizontalFlip(hflip_prob))
            trans.extend([
                T.RandomCrop(crop_size),
                T.ToTensor(),
                T.Normalize(mean=mean, std=std),
            ])
            self.transforms = T.Compose(trans)

        def __call__(self, img, target):
            return self.transforms(img, target)

    class SegPresetTrain:
        def __init__(self, imgsz,crop_size=None, hflip_prob=0.5, mean=(0.485, 0.456, 0.406), std=(0.229, 0.224, 0.225)):

            trans = [T.imgResize((imgsz[0], imgsz[1]))]
            if hflip_prob > 0:
                trans.append(T.RandomHorizontalFlip(hflip_prob))
            # 根据 crop_size 是否存在来决定是否添加 T.RandomCrop
            if crop_size is not None:
                trans.append(T.RandomCrop(crop_size))
            trans.extend([
                T.ToTensor(),
                T.Normalize(mean=mean, std=std),
            ])
            self.transforms = T.Compose(trans)

        def __call__(self, img, target):
            return self.transforms(img, target)
    # class SegPresetEval:
    #     def __init__(self, imgsz, mean=(0.485, 0.456, 0.406), std=(0.229, 0.224, 0.225)):
    #         self.transforms = T.Compose([
    #             T.imgResize((imgsz[0], imgsz[1])),
    #             T.ToTensor(),
    #             T.Normalize(mean=mean, std=std),
    #         ])
    #
    #     def __call__(self, img, target):
    #         return self.transforms(img, target)
    class SegPresetEval:
        def __init__(self, imgsz,crop_size=None, mean=(0.485, 0.456, 0.406), std=(0.229, 0.224, 0.225)):
            # 根据 crop_size 是否存在来决定是否添加 T.RandomCrop
            trans=[T.imgResize((imgsz[0], imgsz[1]))]
            if crop_size is not None:
                trans.append(T.RandomCrop(crop_size))
            trans.extend([
                T.ToTensor(),
                T.Normalize(mean=mean, std=std),
            ])
            self.transforms = T.Compose(trans)

        def __call__(self, img, target):
            return self.transforms(img, target)

    class SegmentationPresetEval:
        def __init__(self, base_size, mean=(0.485, 0.456, 0.406), std=(0.229, 0.224, 0.225)):
            self.transforms = T.Compose([
                T.RandomResize(base_size, base_size),
                T.ToTensor(),
                T.Normalize(mean=mean, std=std),
            ])

        def __call__(self, img, target):
            return self.transforms(img, target)
    def get_transform_seg(self,imgsz,train,crop_size=None):
        return self.SegPresetTrain(imgsz,crop_size=crop_size) if train else self.SegPresetEval(imgsz,crop_size=crop_size)

    def get_transform(self,train):
        base_size = 520
        crop_size = 480

        return self.SegmentationPresetTrain(base_size, crop_size) if train else self.SegmentationPresetEval(base_size)

    def create_model(self,aux, num_classes, pretrain=None):
        # model = fcn_resnet50(aux=aux, num_classes=num_classes)
        if self.params["arch"]=='fcn_resnet50':
            model = fcn_resnet50(aux=aux, num_classes=num_classes)
        ####新增resnet18/34
        elif self.params["arch"]=='fcn_resnet18':
            model = fcn_resnet18(aux=aux, num_classes=num_classes)
        elif self.params["arch"]=='fcn_resnet34':
            model = fcn_resnet34(aux=aux, num_classes=num_classes)
        ####新增resnet18/34
        ####新增deeplabv3_resnet18/34
        elif self.params["arch"]=='deeplabv3_resnet18':
            model = deeplabv3_resnet18(aux=aux, num_classes=num_classes)
        elif self.params["arch"]=='deeplabv3_resnet34':
            model = deeplabv3_resnet34(aux=aux, num_classes=num_classes)
        ####新增deeplabv3_resnet18/34
        else:
            model = models.__dict__[self.params["arch"]](pretrained=False, pretrained_backbone=False, num_classes=num_classes,aux_loss=aux)
        if pretrain:
            # weights_dict = torch.load("./fcn_resnet50_coco.pth", map_location='cpu')
            weights_dict = torch.load(pretrain, map_location='cpu')
            if num_classes != 21:
                # 官方提供的预训练权重是21类(包括背景)
                # 如果训练自己的数据集，将和类别相关的权重删除，防止权重shape不一致报错
                for k in list(weights_dict.keys()):
                    if "classifier.4" in k:
                        del weights_dict[k]
            if "model" in weights_dict:
                ###获取与训练模型最佳分数
                best_miou_allcls = weights_dict['best_miou_allcls']
                if self.params["device"] is not None:
                    # best_acc1 may be from a checkpoint from a different GPU
                    self.best_miou_allcls = best_miou_allcls.to(self.params["device"])
                ###获取与训练模型最佳分数
                missing_keys, unexpected_keys = model.load_state_dict(weights_dict["model"].state_dict(), strict=False)
            else:
                missing_keys, unexpected_keys = model.load_state_dict(weights_dict, strict=False)
            if len(missing_keys) != 0 or len(unexpected_keys) != 0:
                print("missing_keys: ", missing_keys)
                print("unexpected_keys: ", unexpected_keys)

        return model

    def export_model(self,export_modeldict):
        # model = fcn_resnet50(aux=self.aux, num_classes=self.num_classes)
        # weights_dict = torch.load(export_modeldict, map_location='cpu')
        # model.load_state_dict(weights_dict, strict=False)
        # return model
        model=torch.load(export_modeldict, map_location='cpu')['model']
        return model

    def train(self):
        deviceid="cuda:0" if self.params["device"] is None else f"cuda:{self.params['device']}"
        device = torch.device(deviceid if torch.cuda.is_available() else "cpu")
        batch_size = self.params["batch_size"]
        # segmentation nun_classes + background
        crop_size=self.params["crop_size"]

        # with open(self.params["data_yaml_dir"], 'r', encoding='utf-8') as file:
        #     datayaml = yaml.safe_load(file)
        # num_classes = datayaml['nc'] + 1
        # # 提取 train 和 val 路径
        # train_path = datayaml['train']
        # val_path = datayaml['val']
        # # train_dataset = myDataset_yaml(train_path,transforms=get_transform(train=True))
        train_dataset = myDataset_yaml(self.train_path, transforms=self.get_transform_seg(self.params["imgsz"],train=True,crop_size=crop_size))
        # VOCdevkit -> VOC2012 -> ImageSets -> Segmentation -> val.txt
        # val_dataset = myDataset_yaml(val_path,transforms=get_transform(train=False))
        val_dataset = myDataset_yaml(self.val_path, transforms=self.get_transform_seg(self.params["imgsz"],train=False,crop_size=crop_size))

        num_workers = min([os.cpu_count(), batch_size if batch_size > 1 else 0, 8])
        train_loader = torch.utils.data.DataLoader(train_dataset,
                                                   batch_size=batch_size,
                                                   num_workers=num_workers,
                                                   shuffle=True,
                                                   pin_memory=True,
                                                   collate_fn=train_dataset.collate_fn)

        val_loader = torch.utils.data.DataLoader(val_dataset,
                                                 batch_size=1,
                                                 num_workers=num_workers,
                                                 pin_memory=True,
                                                 collate_fn=val_dataset.collate_fn)

        model = self.create_model(self.params["aux"], self.num_classes,self.params["pretrained_model"])
        model.to(device)

        params_to_optimize = [
            {"params": [p for p in model.backbone.parameters() if p.requires_grad]},
            {"params": [p for p in model.classifier.parameters() if p.requires_grad]}
        ]

        if self.params["aux"]:
            params = [p for p in model.aux_classifier.parameters() if p.requires_grad]
            params_to_optimize.append({"params": params, "lr": self.params["lr"] * 10})

        optimizer = torch.optim.SGD(
            params_to_optimize,
            lr=self.params["lr"], momentum=self.params["momentum"], weight_decay=self.params["weight_decay"]
        )

        scaler = torch.cuda.amp.GradScaler() if self.params["amp"] else None

        # 创建学习率更新策略，这里是每个step更新一次(不是每个epoch)
        lr_scheduler = create_lr_scheduler(optimizer, len(train_loader), self.params["epochs"], warmup=True)
        start_epoch=0
        if self.params["resume"]:
            checkpoint = torch.load(self.params["pretrained_model"], map_location='cpu')
            model.load_state_dict(checkpoint['model'].state_dict())
            optimizer.load_state_dict(checkpoint['optimizer'])
            lr_scheduler.load_state_dict(checkpoint['lr_scheduler'])
            start_epoch = checkpoint['epoch'] + 1
            if self.params["amp"]:
                scaler.load_state_dict(checkpoint["scaler"])
            train_dir = checkpoint['outdir']
            epochs_all=checkpoint['epoch_all']
        else:
            # 用来保存训练以及验证过程中信息
            ####创建训练输出子文件train####
            from ultralytics.utils.files import increment_path
            train_childdir=os.path.join(self.params["save_dir"],'FCNtrain')
            train_dir = str(increment_path(train_childdir, exist_ok=False)).replace("\\",'/')
            ######
            if not os.path.exists(self.params["save_dir"]):
                os.mkdir(self.params["save_dir"])
            if not os.path.exists(train_dir):
                os.mkdir(train_dir)
            epochs_all = self.params["epochs"]

        # results_file = train_dir+"/results{}.txt".format(datetime.datetime.now().strftime("%Y%m%d-%H%M%S"))
        results_file = train_dir + "/results.txt"
        # ###读取train、val的images路径

        start_time = time.time()
        for epoch in range(start_epoch, epochs_all):
            mean_loss, lr = train_one_epoch(model, optimizer, train_loader, device, epoch,
                                            lr_scheduler=lr_scheduler, print_freq=self.params["traininfo_printepoch"], scaler=scaler)

            confmat ,miou_allcls= evaluate(model, val_loader, device=device, num_classes=self.num_classes)
            #判断是否为best——model
            is_best = miou_allcls > self.best_miou_allcls
            self.best_miou_allcls = max(miou_allcls,self.best_miou_allcls)
            # 判断是否为best——model
            val_info = str(confmat)
            print(val_info)
            # write into txt
            with open(results_file, "a") as f:
                # 记录每个epoch对应的train_loss、lr以及验证集各指标
                train_info = f"[epoch: {epoch}]\n" \
                             f"train_loss: {mean_loss:.4f}\n" \
                             f"lr: {lr:.6f}\n"
                f.write(train_info + val_info + "\n\n")
            #原来的save_file 里"model": model.state_dict(),现改为model
            save_file = {"model": model,
                         "optimizer": optimizer.state_dict(),
                         "lr_scheduler": lr_scheduler.state_dict(),
                         "epoch": epoch,
                         "epoch_all": self.params["epochs"],
                         "outdir":train_dir,
                         "best_miou_allcls": self.best_miou_allcls,
                         "num_classes": self.num_classes - 1,
                         "model_type": self.params["arch"],
                         "args": self.params}
            if self.params["amp"]:
                save_file["scaler"] = scaler.state_dict()
            if not os.path.exists(train_dir+'/weights'):
                os.mkdir(train_dir+'/weights')
            # fcnmodel_savedir=train_dir+'/weights'+"/model_{}.pth".format(epoch)
            # self.model_savedir=fcnmodel_savedir
            # torch.save(save_file, fcnmodel_savedir)
            fcnmodel_savedir=train_dir+'/weights'
            lastpath = fcnmodel_savedir + "/last.pt"
            bestpath = fcnmodel_savedir + "/best.pt"
            torch.save(save_file, lastpath)
            if is_best or (not os.path.exists(bestpath)):
                shutil.copyfile(lastpath, bestpath)
            self.model_savedir=bestpath

        total_time = time.time() - start_time
        total_time_str = str(datetime.timedelta(seconds=int(total_time)))
        print("training time {}".format(total_time_str))


#


# if __name__ == '__main__':
#     args = parse_args()
#
#     if not os.path.exists("./save_weights"):
#         os.mkdir("./save_weights")
#
#     main(args)



import os
import yaml
from PIL import Image
import numpy as np

class HSAHI_det:
    def __init__(self, yaml_path=None,slice_height=256, slice_width=256, overlap_height_ratio=0.25, overlap_width_ratio=0.25,sahi_datadir=None,roi=None):
        self.slice_height_ = slice_height
        self.slice_width_ = slice_width
        self.overlap_height_ratio_ = overlap_height_ratio
        self.overlap_width_ratio_ = overlap_width_ratio
        self.yaml_path=yaml_path
        self.output_yaml_path=None
        self.slice_regions = []
        self.Data_sahi_dir=sahi_datadir
        self.roi=[roi[0],roi[1],roi[2]-roi[0],roi[3]-roi[1]] if roi else roi
    def calculate_slice_regions(self, image_height, image_width,roi=None):
        #是否有roi，有的话使用
        if roi:
            image_width=roi[2]
            image_height=roi[3]
        # 是否有roi，有的话使用
        self.slice_regions = []
        step_height = self.slice_height_ - int(self.slice_height_ * self.overlap_height_ratio_)
        step_width = self.slice_width_ - int(self.slice_width_ * self.overlap_width_ratio_)
        index = 0

        # 如果图像的高度或宽度小于切图尺寸，调整切图尺寸
        if image_height < self.slice_height_:
            self.slice_height_ = image_height
        if image_width < self.slice_width_:
            self.slice_width_ = image_width

        # 使用一个集合来跟踪已添加的切片区域
        seen_regions = set()

        for y in range(0, image_height, step_height):
            for x in range(0, image_width, step_width):
                width = self.slice_width_
                height = self.slice_height_

                temp_x = x
                temp_y = y

                if x + width > image_width:
                    temp_x -= (x + width) - image_width
                if y + height > image_height:
                    temp_y -= (y + height) - image_height

                # 创建当前切片区域
                # 是否有roi，有的话使用
                if roi:
                    region = (temp_x+roi[0], temp_y+roi[1], width, height)
                else:
                    region = (temp_x, temp_y, width, height)
                # 使用元组作为唯一标识符，添加到集合中
                if region not in seen_regions:
                    seen_regions.add(region)
                    self.slice_regions.append((region, index))
                    index += 1
###获取切图区域内的标注框####
    def bbox_in_slice(self, bbox, slice_region):
        slice_x, slice_y, slice_w, slice_h = slice_region
        bbox_x1, bbox_y1, bbox_x2, bbox_y2 = bbox

        new_bbox_x1 = max(bbox_x1 - slice_x, 0)
        new_bbox_y1 = max(bbox_y1 - slice_y, 0)
        new_bbox_x2 = min(bbox_x2 - slice_x, slice_w)
        new_bbox_y2 = min(bbox_y2 - slice_y, slice_h)

        if new_bbox_x1 < new_bbox_x2 and new_bbox_y1 < new_bbox_y2:
            return new_bbox_x1, new_bbox_y1, new_bbox_x2, new_bbox_y2
        else:
            return None

    def convert_to_original(self, size, box):
        width, height = size
        x = box[0] * width
        y = box[1] * height
        w = box[2] * width
        h = box[3] * height

        x1 = x - w / 2
        y1 = y - h / 2
        x2 = x + w / 2
        y2 = y + h / 2

        return x1, y1, x2, y2
    def process_imgs(self, image_paths, output_dir):
        output_folder = os.path.join(output_dir, 'images')
        txt_outdir=os.path.join(output_dir,'labels')
        if not os.path.exists(output_folder):
            os.makedirs(output_folder)
        if not os.path.exists(txt_outdir):
            os.makedirs(txt_outdir)
        new_image_paths = []  # 用于存储新的图片路径

        for image_path in image_paths:
            ##获取图片格式
            imgtype = os.path.basename(image_path).split('.')[-1]
            txt_path = image_path.replace(f".{imgtype}", ".txt").replace("images", "labels")
            with open(txt_path, "r") as f:
                annotations = f.readlines()

            with Image.open(image_path) as img:
                img.load()
                width, height = img.size
                self.calculate_slice_regions(image_height=height, image_width=width,roi=self.roi)

                for slice_region, index in self.slice_regions:
                    slice_x, slice_y, slice_w, slice_h = slice_region
                    slice_img = img.crop((slice_x, slice_y, slice_x + slice_w, slice_y + slice_h))

                    slice_img_filename = os.path.join(output_folder, f"{os.path.basename(image_path).replace(f'.{imgtype}', '')}_slice_{index}.{imgtype}")
                    slice_img.save(slice_img_filename)

                    slice_txt_filename = os.path.join(txt_outdir, f"{os.path.basename(image_path).replace(f'.{imgtype}', '')}_slice_{index}.txt")
                    with open(slice_txt_filename, "w") as slice_f:
                        has_bbox = False
                        for annotation in annotations:
                            elements = annotation.strip().split()
                            bbox_id = int(elements[0])
                            normalized_bbox = list(map(float, elements[1:]))

                            bbox = self.convert_to_original((width, height), normalized_bbox)

                            new_bbox = self.bbox_in_slice(bbox, slice_region)
                            ###归一化标注框
                            if new_bbox:
                                has_bbox = True
                                x_center = (new_bbox[0] + new_bbox[2]) / 2 / slice_w
                                y_center = (new_bbox[1] + new_bbox[3]) / 2 / slice_h
                                bbox_w = (new_bbox[2] - new_bbox[0]) / slice_w
                                bbox_h = (new_bbox[3] - new_bbox[1]) / slice_h

                                slice_f.write(f"{bbox_id} {x_center} {y_center} {bbox_w} {bbox_h}\n")

                        if not has_bbox:
                            slice_f.write("")

                    new_image_paths.append(slice_img_filename)  # 添加新的图片路径

        return new_image_paths
    def update_yaml(self, train_paths, val_paths,test_paths):
        with open(self.yaml_path, 'r', encoding='utf-8') as file:
            data = yaml.safe_load(file)
        train_paths = train_paths.replace("\\", "/")
        val_paths = val_paths.replace("\\", "/")
        test_paths = test_paths.replace("\\", "/")
        data['train'] = train_paths
        data['val'] = val_paths
        data['test'] = test_paths
        with open(self.output_yaml_path, 'w', encoding='utf-8') as file:
            yaml.dump(data, file, allow_unicode=True)

    def save_image_paths(self, image_paths, output_txt_path):
        with open(output_txt_path, 'w') as f:
            for path in image_paths:
                f.write(f"{path}\n")
    def read_yaml(self):
        with open(self.yaml_path, 'r', encoding='utf-8') as file:
            config = yaml.safe_load(file)

        train_txt_path = config['train']
        val_txt_path = config['val']
        test_txt_path=train_txt_path.replace("train.txt",'test.txt')
        # 3. 读取 train 和 val 的文件路径
        train_image_paths = [line.strip() for line in open(train_txt_path, 'r')]
        val_image_paths = [line.strip() for line in open(val_txt_path, 'r')]
        # 检查 test.txt 文件是否存在
        if os.path.exists(test_txt_path):
            test_image_paths = [line.strip() for line in open(test_txt_path, 'r')]
        else:
            # 如果 test.txt 文件不存在，则返回空列表
            test_image_paths = []
            print(f"Warning: {test_txt_path} does not exist. Returning empty test image paths.")
        return train_image_paths, val_image_paths, test_image_paths


    def process_data(self):
        # 5. 处理 train 和 val 图片
        train_image_paths, val_image_paths, test_image_paths=self.read_yaml()
        if self.Data_sahi_dir:
            Data_sahi_dir=self.Data_sahi_dir
        else:
        # 获取上层路径
            parent_dir = os.path.dirname(self.yaml_path)
            Data_sahi_dir=os.path.join(parent_dir,'Data_det_sahi')
        if not os.path.exists(Data_sahi_dir):
            os.makedirs(Data_sahi_dir)
        trainval_outdir=os.path.join(Data_sahi_dir,'paper_data')
        if not os.path.exists(trainval_outdir):
            os.makedirs(trainval_outdir)
        new_train_paths = self.process_imgs(train_image_paths,Data_sahi_dir)
        new_val_paths = self.process_imgs(val_image_paths, Data_sahi_dir)
        #
        # ###和train、val一样使用切图切分，然后输出切图后的路径
        # new_test_paths = self.process_imgs(test_image_paths, Data_sahi_dir)
        # ###对test数据集不使用切图，直接给出原来一样的路径
        new_test_paths =test_image_paths
        #
        # 6. 保存新的 train.txt 和 val.txt 文件
        output_train_txt=os.path.join(trainval_outdir,'train.txt')
        output_val_txt=os.path.join(trainval_outdir,'val.txt')
        output_test_txt = os.path.join(trainval_outdir, 'test.txt')
        self.save_image_paths(new_train_paths, output_train_txt)
        self.save_image_paths(new_val_paths, output_val_txt)
        self.save_image_paths(new_test_paths, output_test_txt)
        # 7. 更新 YAML 文件
        self.output_yaml_path=os.path.join(Data_sahi_dir,'mydata.yaml')
        self.update_yaml(output_train_txt, output_val_txt,output_test_txt)

    ##predict_slice_imgs切分文件夹所有图片
    def predict_slice_imgs(self, image_paths):
        sliced_images = []  # 存储切割后的图片数据
        for image_path in os.listdir(image_paths):
            # 获取图片格式
            sliced_image=[]
            image_path = os.path.join(image_paths,image_path)
            with Image.open(image_path) as img:
                width, height = img.size
                self.calculate_slice_regions(image_height=height, image_width=width,roi=self.roi)

                for slice_region, index in self.slice_regions:
                    slice_x, slice_y, slice_w, slice_h = slice_region
                    slice_img = img.crop((slice_x, slice_y, slice_x + slice_w, slice_y + slice_h))

                    # 将切割的图像保存为 numpy 数组（或直接传递到模型支持的格式）
                    slice_img_array = np.array(slice_img)
                    sliced_image.append(slice_img_array)
            sliced_images.append(sliced_image)
        # 返回所有切割后的图片列表
        return sliced_images

    ##predict_slice_oneimg切分单个文件图片
    def predict_slice_oneimg(self, image_path_or_image):
        # 根据参数的类型来决定如何处理
        if isinstance(image_path_or_image, str):
            sliced_images = []  # 存储切割后的图片数据
            image_path=image_path_or_image
            with Image.open(image_path) as img:
                width, height = img.size
                self.calculate_slice_regions(image_height=height, image_width=width,roi=self.roi)

                for slice_region, index in self.slice_regions:
                    slice_x, slice_y, slice_w, slice_h = slice_region
                    slice_img = img.crop((slice_x, slice_y, slice_x + slice_w, slice_y + slice_h))

                    # 将切割的图像保存为 numpy 数组（或直接传递到模型支持的格式）
                    slice_img_array = np.array(slice_img)
                    sliced_images.append(slice_img_array)
            # 返回所有切割后的图片列表
            return sliced_images
        elif isinstance(image_path_or_image, Image.Image):
            sliced_images = []  # 存储切割后的图片数据
            img = image_path_or_image
            width, height = img.size
            self.calculate_slice_regions(image_height=height, image_width=width,roi=self.roi)

            for slice_region, index in self.slice_regions:
                slice_x, slice_y, slice_w, slice_h = slice_region
                slice_img = img.crop((slice_x, slice_y, slice_x + slice_w, slice_y + slice_h))

                # 将切割的图像保存为 numpy 数组（或直接传递到模型支持的格式）
                slice_img_array = np.array(slice_img)
                sliced_images.append(slice_img_array)
            # 返回所有切割后的图片列表
            return sliced_images
        else:
            raise TypeError("image_path_or_image 必须是字符串或 PIL.Image.Image")

# # 1. 初始化 SAHI 类
# slice_height = 256
# slice_width = 256
# overlap_height_ratio = 0.2
# overlap_width_ratio = 0.2
#
# sahi = SAHI(slice_height, slice_width, overlap_height_ratio, overlap_width_ratio)

# 2. 读取 YAML 文件
# yaml_path = r"E:\ALLvision\pycharmproject\Efficientnet\segjjc_clsseg_try\sahi_preprocess\1014\try\Data_det\mydata.yaml"
# with open(yaml_path, 'r', encoding='utf-8') as file:
#     config = yaml.safe_load(file)
#
# train_txt_path = config['train']
# val_txt_path = config['val']
#
# # 3. 读取 train 和 val 的文件路径
# train_image_paths = [line.strip() for line in open(train_txt_path, 'r')]
# val_image_paths = [line.strip() for line in open(val_txt_path, 'r')]

# # 4. 指定输出文件夹
# output_image_folder = r"E:\ALLvision\pycharmproject\Efficientnet\segjjc_clsseg_try\sahi_preprocess\1014\try\out\imgs"
# output_train_txt = r"E:\ALLvision\pycharmproject\Efficientnet\segjjc_clsseg_try\sahi_preprocess\1014\try\out\paper_data\train.txt"
# output_val_txt = r"E:\ALLvision\pycharmproject\Efficientnet\segjjc_clsseg_try\sahi_preprocess\1014\try\out\paper_data\val.txt"
# output_yaml_path = r"E:\ALLvision\pycharmproject\Efficientnet\segjjc_clsseg_try\sahi_preprocess\1014\try\out\output_config.yaml"

# # 5. 处理 train 和 val 图片
# new_train_paths = sahi.process_txt_files(train_image_paths, os.path.dirname(train_txt_path), output_image_folder)
# new_val_paths = sahi.process_txt_files(val_image_paths, os.path.dirname(val_txt_path), output_image_folder)
#
# # 6. 保存新的 train.txt 和 val.txt 文件
# sahi.save_image_paths(new_train_paths, output_train_txt)
# sahi.save_image_paths(new_val_paths, output_val_txt)
#
# # 7. 更新 YAML 文件
# sahi.update_yaml(yaml_path, output_train_txt, output_val_txt, output_yaml_path)
# yamlpath=r'E:\ALLvision\pycharmproject\Efficientnet\segjjc_clsseg_try\sahi_preprocess\1014\try\Data_pest\mydata2\mydata.yaml'
# sahi=SAHI(yamlpath)
# sahi.process_data()
# print("处理完成！")



import torch
import torch.nn as nn
import torchvision.models as models
from torchvision.models._utils import IntermediateLayerGetter
from torchvision.models.segmentation.deeplabv3 import DeepLabV3_ResNet101_Weights,DeepLabV3, DeepLabHead, FCNHead
from torchvision.models.resnet import ResNet18_Weights, ResNet34_Weights
from typing import Optional, Any


def _deeplabv3_resnet_custom(
        backbone: models.ResNet,
        num_classes: int,
        aux: Optional[bool],
) -> DeepLabV3:
    """
    适配 ResNet-18/34 作为 DeepLabV3 的 backbone。
    """
    return_layers = {"layer4": "out"}
    if aux:
        return_layers["layer3"] = "aux"

    backbone = IntermediateLayerGetter(backbone, return_layers=return_layers)

    # 适配 ResNet-18/34，layer4 输出通道数是 512，而 ResNet-50/101 是 2048
    classifier = DeepLabHead(512, num_classes)
    aux_classifier = FCNHead(256, num_classes) if aux else None

    return DeepLabV3(backbone, classifier, aux_classifier)

def deeplabv3_resnet18(
    *,
    weights: Optional[DeepLabV3_ResNet101_Weights] = None,
    progress: bool = True,
    num_classes: Optional[int] = None,
    aux: Optional[bool] = None,
    weights_backbone: Optional[ResNet18_Weights] = None,  # 不使用预训练权重
    **kwargs: Any,
) -> DeepLabV3:
    """Constructs a DeepLabV3 model with a ResNet-18 backbone (no pretrained weights)."""

    if weights is not None:
        weights_backbone = None
        num_classes = _overwrite_value_param("num_classes", num_classes, len(weights.meta["categories"]))
        aux = _overwrite_value_param("aux_loss", aux, True)
    elif num_classes is None:
        num_classes = 21

    # 不加载预训练权重
    backbone = models.resnet18(weights=weights_backbone, replace_stride_with_dilation=[False, False, False])
    model = _deeplabv3_resnet_custom(backbone, num_classes, aux)

    if weights is not None:
        model.load_state_dict(weights.get_state_dict(progress=progress))

    return model

def deeplabv3_resnet34(
    *,
    weights: Optional[DeepLabV3_ResNet101_Weights] = None,
    progress: bool = True,
    num_classes: Optional[int] = None,
    aux: Optional[bool] = None,
    weights_backbone: Optional[ResNet34_Weights] = None,  # 不使用预训练权重
    **kwargs: Any,
) ->DeepLabV3:
    """Constructs a DeepLabV3 model with a ResNet-34 backbone (no pretrained weights)."""

    if weights is not None:
        weights_backbone = None
        num_classes = _overwrite_value_param("num_classes", num_classes, len(weights.meta["categories"]))
        aux = _overwrite_value_param("aux_loss", aux, True)
    elif num_classes is None:
        num_classes = 21

    # 不加载预训练权重
    backbone = models.resnet34(weights=weights_backbone, replace_stride_with_dilation=[False, False, False])
    model = _deeplabv3_resnet_custom(backbone, num_classes, aux)

    if weights is not None:
        model.load_state_dict(weights.get_state_dict(progress=progress))

    return model

def _overwrite_value_param(param_name, param_value, default_value):
    """Helper function to overwrite parameter values if needed."""
    return default_value if param_value is None else param_value



import os
import time
import json

import torch
from torchvision import transforms
import numpy as np
from PIL import Image
from transforms import letterbox_pil
from src import fcn_resnet50
from typing import Tuple

def time_synchronized():
    torch.cuda.synchronize() if torch.cuda.is_available() else None
    return time.time()

class imgResize_solo(object):
    def __init__(self,size:Tuple[int, int]):
        self.max_size = (max(size),max(size))

    def __call__(self, image):
        image=letterbox_pil(image,self.max_size)
        return image

def main():
    aux = False  # inference time not need aux_classifier
    classes = 3
    weights_path = "./save_weights/model_19.pth"
    img_path = "./testdog/1_horizontal.jpg"
    palette_path = "./palette.json"
    assert os.path.exists(weights_path), f"weights {weights_path} not found."
    assert os.path.exists(img_path), f"image {img_path} not found."
    assert os.path.exists(palette_path), f"palette {palette_path} not found."
    with open(palette_path, "rb") as f:
        pallette_dict = json.load(f)
        pallette = []
        for v in pallette_dict.values():
            pallette += v

    # get devices
    device = torch.device("cuda:0" if torch.cuda.is_available() else "cpu")
    print("using {} device.".format(device))

    # create model
    model = fcn_resnet50(aux=aux, num_classes=classes+1)

    # delete weights about aux_classifier
    weights_dict = torch.load(weights_path, map_location='cpu')['model']
    for k in list(weights_dict.keys()):
        if "aux" in k:
            del weights_dict[k]

    # load weights
    model.load_state_dict(weights_dict)
    model.to(device)

    # load image
    original_img = Image.open(img_path)

    # from pil image to tensor and normalize
    data_transform = transforms.Compose([imgResize_solo((640, 640)),
                                         transforms.ToTensor(),
                                         transforms.Normalize(mean=(0.485, 0.456, 0.406),
                                                              std=(0.229, 0.224, 0.225))])
    img = data_transform(original_img)
    # expand batch dimension
    img = torch.unsqueeze(img, dim=0)

    model.eval()  # 进入验证模式
    with torch.no_grad():
        # init model
        img_height, img_width = img.shape[-2:]
        init_img = torch.zeros((1, 3, img_height, img_width), device=device)
        model(init_img)

        t_start = time_synchronized()
        output = model(img.to(device))
        t_end = time_synchronized()
        print("inference time: {}".format(t_end - t_start))

        prediction = output['out'].argmax(1).squeeze(0)
        prediction = prediction.to("cpu").numpy().astype(np.uint8)
        mask = Image.fromarray(prediction)
        mask.save("E:/ALLvision/pycharmproject/yoloseg/githubtry/diyFCN/try2_yaml/fcn/testdog/test_result.png")
        mask.putpalette(pallette)
        # mask.save("test_result.png")


if __name__ == '__main__':
    main()



from setuptools import setup, find_packages

setup(
    name="SegJJC",
    version="0.0.12",
    author="hfh",
    author_email="your.email@example.com",
    description="A sample Python package with a .pyd file",
    url="https://gitee.com/hfh936784552/Segjjc_git",
    packages=find_packages(),
    include_package_data=True,
    package_data={
        "SegJJC": ["dll/*.dll",
                   "fcn/*.json",
                   "fcn/*.txt",
                   "demo/*.json",],  # 指定要包含的 .pyd 文件
    },
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
    ],
    python_requires=">=3.6",
    install_requires=[],
)



# from setuptools import setup, find_packages
# from glob import glob
# import os
#
# # # 查找所有的 .pyd 文件
# # pyd_files = glob("my_package/**/*.pyd", recursive=True)
#
# setup(
#     name="netpin2excel",                # 替换为你的包名
#     version="0.1.1",                  # 版本号
#     author="hfh",               # 作者信息
#     author_email="your.email@example.com",
#     description="A sample Python package with a .pyd file",
#     url="https://github.com/yourusername/my_package",  # 项目主页
#     packages=find_packages(),         # 自动查找包
#     include_package_data=True,        # 包含所有文件，包括 .pyd 文件
#     # package_data={
#     #     "my_package": ["netpin_process.pyd"],      # 指定包含 .pyd 文件
#     # },
#     classifiers=[
#         "Programming Language :: Python :: 3",
#         "License :: OSI Approved :: MIT License",
#         "Operating System :: OS Independent",
#     ],
#     python_requires=">=3.6",          # Python 版本要求
#     install_requires=[],              # 依赖项，可根据需要添加
# )



import torch
import torchvision.models.segmentation as models

from ..fcn.src import *

class export_model:

    @staticmethod
    def export_model_efficient(export_modeldict):
        model = torch.load(export_modeldict, map_location='cpu')['state_dict']
        return model

    @staticmethod
    def export_model_fcn(export_modeldict):
        # model = fcn_resnet50(aux=self.aux, num_classes=self.num_classes)
        # weights_dict = torch.load(export_modeldict, map_location='cpu')
        # model.load_state_dict(weights_dict, strict=False)
        # return model
        # 加载模型权重
        model_pt= torch.load(export_modeldict, map_location='cpu')
        diymodel_dict = {
            "fcn_resnet50": fcn_resnet50,
            "fcn_resnet18": fcn_resnet18,
            "fcn_resnet34": fcn_resnet34,
            "deeplabv3_resnet18": deeplabv3_resnet18,
            "deeplabv3_resnet34": deeplabv3_resnet34,
        }

        model_arch = model_pt["model_type"]
        if model_arch in diymodel_dict:
            model = diymodel_dict[model_arch](aux=False, num_classes=model_pt['num_classes'] + 1)
        else:
            model = models.__dict__[model_arch](pretrained=False, pretrained_backbone=False,
                                                      num_classes=model_pt['num_classes'] + 1,
                                                      aux_loss=False)
        model_dict = model_pt['model']
        # 只保留去除了 aux 相关参数的 state_dict
        weights_dict = model_dict.state_dict()

        # 删除包含 "aux" 关键字的权重
        for k in list(weights_dict.keys()):
            if "aux" in k:
                del weights_dict[k]

        # 重新包装 model，返回修改后的状态字典
        model.load_state_dict(weights_dict)
        return model


import os
import cv2
import numpy as np
from PIL import Image


def palette():
    # 获取当前目录
    root = os.getcwd()
    # 灰度图的位置
    imgFile = root + '\\SegmentationClass'
    # 将所有的灰度图添加调色板
    for i, img in enumerate(os.listdir(imgFile)):
        filename, _ = os.path.splitext(img)
        # 这个地方需要将图片路径添加完整，不然后面读取图片文件不存在
        img = 'SegmentationClass/' + img
        img = cv2.imread(img, cv2.IMREAD_GRAYSCALE)
        save_path = root + '\\imgPalette\\' + filename + '.png'
        img = Image.fromarray(img)  # 将图像从numpy的数据格式转为PIL中的图像格式
        palette = []
        for i in range(256):
            palette.extend((i, i, i))
        # 这里设置21个颜色，其中背景为黑色，总共21个类别（包括背景）
        palette[:3 * 21] = np.array([[0, 0, 0], [0, 255, 255], [0, 128, 0], [128, 128, 0], [0, 0, 128],
                                     [128, 0, 128], [0, 128, 128], [128, 128, 128], [64, 0, 0], [192, 0, 0],
                                     [64, 128, 0], [192, 128, 0], [64, 0, 128], [192, 0, 128], [64, 128, 128],
                                     [192, 128, 128], [0, 64, 0], [128, 64, 0], [0, 192, 0], [128, 192, 0], [0, 64, 128]
                                     ], dtype='uint8').flatten()

        img.putpalette(palette)
        # print(np.shape(palette)) 输出（768,)
        img.save(save_path)


if __name__ == '__main__':
    print('Pycharm')
    palette()